from .base import CTERAException


class AuthenticationError(CTERAException):
    """
    Exception raised for authentication failures.
    """

    def __init__(self):
        super().__init__("Authentication failed: Invalid username or password")



import logging

from ..exceptions.transport import InternalServerError
from ..exceptions.auth import AuthenticationError
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Login(BaseCommand):

    def info(self):
        """
        Get login info
        """
        return self._edge.api.get('/nosession/logininfo')

    def login(self, username, password):
        """
        Log in to CTERA Edge Filer

        :param str username: User name
        :param str password: User password
        :raises: :class:`cterasdk.exceptions.auth.AuthenticationError`
        """
        host = self._edge.host()
        try:
            self._edge.api.form_data('/login', {'username': username, 'password': password})
            logger.info("User logged in. %s", {'host': host, 'user': username})
            self._edge.ctera_migrate.login()
        except InternalServerError as e:
            if e.error.response.error.msg == 'Wrong username or password':
                raise AuthenticationError() from e
            raise

    def sso(self, ticket):
        """
        Single Sign On

        :param str ticket: SSO Ticket.
        """
        logger.info("Performing Single Sign On.")
        self._edge.api.get('/ssologin', params={'ticket': ticket})
        self._edge.ctera_migrate.login()

    def logout(self):
        host = self._edge.host()
        user = self._edge.session().account.name
        self._edge.api.form_data('/logout', {'foo': 'bar'})
        logger.info("User logged out. %s", {'host': host, 'user': user})



class ConnectionStatus:
    """Connection Status"""

    Connected = 'connected'
    Connecting = 'connecting'
    Disconnected = 'disconnected'


class Product:
    """Product"""

    Portal = 'Portal'
    Edge = 'Edge'
    Drive = 'Drive'



import logging
from ..lib.tasks import AwaitablePortalTask
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.core')


class Tasks(BaseCommand):
    """ Portal Background Task APIs """

    def awaitable_task(self, ref):
        return AwaitablePortalTask(self._core, ref)

    def wait(self, ref, timeout=None, poll_interval=None):
        awaitable_task = AwaitablePortalTask(self._core, ref)
        return awaitable_task.wait(timeout, poll_interval)



import cterasdk.settings
from ..services import AsyncManagement
from ..endpoints import EndpointBuilder
from ...clients import clients
from .. import authenticators
from ...lib.session.core import Session
from ...asynchronous.core import files, login, cloudfs, notifications, portals, settings, tasks, users


class Clients:

    def __init__(self, core):
        self.v1 = V1(core)
        self.v2 = V2(core)
        self.io = IO(core)


class V1:

    def __init__(self, core):
        self.ctera = core.default.clone(clients.AsyncExtended, EndpointBuilder.new(core.base, core.context))
        self.api = core.default.clone(clients.AsyncAPI, EndpointBuilder.new(core.base, core.context, '/api'))


class V2:

    def __init__(self, core):
        self.api = core.default.clone(clients.AsyncJSON, EndpointBuilder.new(core.base, core.context, '/v2/api'))


class IO:

    def __init__(self, core):
        self._folders = core.default.clone(clients.AsyncFolders, EndpointBuilder.new(core.base, core.context, '/folders/folders'))
        self._upload = core.default.clone(clients.AsyncUpload, EndpointBuilder.new(core.base, core.context, '/upload/folders'))
        self._webdav = core.default.clone(clients.AsyncWebDAV, EndpointBuilder.new(core.base, core.context, '/webdav'))

    @property
    def upload(self):
        return self._upload.upload

    @property
    def download(self):
        return self._webdav.get

    @property
    def download_zip(self):
        return self._folders.download_zip

    @property
    def builder(self):
        return self._webdav._builder  # pylint: disable=protected-access


class AsyncPortal(AsyncManagement):

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https, None, cterasdk.settings.core.asyn.settings)
        self._ctera_session = Session(self.host(), self.context)
        self._ctera_clients = Clients(self)
        self.cloudfs = cloudfs.CloudFS(self)
        self.files = files.CloudDrive(self)
        self.notifications = notifications.Notifications(self)
        self.settings = settings.Settings(self)
        self.tasks = tasks.Tasks(self)
        self.users = users.Users(self)

    @property
    def v1(self):
        return self.clients.v1

    @property
    def v2(self):
        return self.clients.v2

    @property
    def io(self):
        return self.clients.io

    @property
    def _login_object(self):
        return login.Login(self)

    def _authenticator(self, url):
        return authenticators.core(self.session(), url, self.context)


class AsyncGlobalAdmin(AsyncPortal):

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https)
        self.portals = portals.Portals(self)

    @property
    def context(self):
        return 'admin'

    async def impersonate(self, username, tenant):
        """
        Impersonate a Portal user

        :param str username: Username
        :param str tenant: Tenant
        """
        ctera_ticket = await self.users.generate_ticket(username, tenant)
        user = AsyncServicesPortal(f'{tenant}.{await self.settings.global_settings.dns_suffix}', self.port())
        await user.sso(ctera_ticket)
        return user


class AsyncServicesPortal(AsyncPortal):

    @property
    def context(self):
        return 'ServicesPortal'

    async def sso(self, ctera_ticket):
        """
        Login using a Portal ticket

        :param str ctera_ticket: SSO Ticket.
        """
        await self._login_object.sso(ctera_ticket)
        await self._ctera_session.async_start_session(self)
        await self.v1.api.web_session()



from abc import ABC, abstractmethod
from http import HTTPStatus
from ..exceptions.transport import (
    BadRequest, Unauthorized, Forbidden, NotFound, NotAllowed, PreConditionFailed, Unprocessable,
    InternalServerError, BadGateway, ServiceUnavailable, GatewayTimeout, HTTPError
)
from ..common import Object
from ..convert import fromjsonstr, fromxmlstr


class Error(Object):
    """
    Error object.

    :ivar cterasdk.common.object.Object request: Request
    :ivar cterasdk.common.object.Object response: Response
    """
    def __init__(self, response, error):
        super().__init__(
            request=Object(
                method=response.method,
                url=response.real_url.human_repr()
            ),
            response=Object(
                status=response.status,
                error=error
            )
        )


class ErrorHandler(ABC):

    async def a_accept(self, response):
        error = self._accept(response, await response.text())
        raise_error(response.status, error)

    def accept(self, response):
        error = self._accept(response, response.text())
        raise_error(response.status, error)

    @abstractmethod
    def _accept(self, response, message):
        raise NotImplementedError("Subclass must implement the '_accept' method")


class DefaultHandler(ErrorHandler):

    def _accept(self, response, message):
        return Error(response, message)


class XMLHandler(ErrorHandler):

    def _accept(self, response, message):
        return Error(response, fromxmlstr(message) or message)


class JSONHandler(ErrorHandler):

    def _accept(self, response, message):
        return Error(response, fromjsonstr(message) or message)


def raise_error(status, error):
    exceptions = {
        HTTPStatus.BAD_REQUEST: BadRequest,
        HTTPStatus.UNAUTHORIZED: Unauthorized,
        HTTPStatus.FORBIDDEN: Forbidden,
        HTTPStatus.NOT_FOUND: NotFound,
        HTTPStatus.METHOD_NOT_ALLOWED: NotAllowed,
        HTTPStatus.PRECONDITION_FAILED: PreConditionFailed,
        HTTPStatus.UNPROCESSABLE_ENTITY: Unprocessable,
        HTTPStatus.INTERNAL_SERVER_ERROR: InternalServerError,
        HTTPStatus.BAD_GATEWAY: BadGateway,
        HTTPStatus.SERVICE_UNAVAILABLE: ServiceUnavailable,
        HTTPStatus.GATEWAY_TIMEOUT: GatewayTimeout
    }
    exception = exceptions.get(status, HTTPError)
    raise exception(error)



import re
import logging
from contextlib import contextmanager
from ..objects.uri import quote, unquote
from ..common import Object, DateTimeUtils
from ..core.enum import ProtectionLevel, CollaboratorType, SearchType, PortalAccountType, FileAccessMode, \
    UploadError, ResourceAction, ResourceScope
from ..core.types import PortalAccount, UserAccount, GroupAccount
from ..exceptions.io import ResourceExistsError, PathValidationError, NameSyntaxError, \
    ReservedNameError, RestrictedRoot, PermissionDenied, FileConflict, RemoteStorageError, NotADirectory, \
    UnwriteableScope
from ..exceptions.io import UploadException, OutOfQuota, RejectedByPolicy, NoStorageBucket, WindowsACLError
from ..lib.iterator import DefaultResponse
from . import common


logger = logging.getLogger('cterasdk.core')


class CorePath(common.BasePath):
    """Path for CTERA Portal"""

    def __init__(self, scope, reference):
        """
        Initialize a CTERA Portal Path.

        :param str scope: Scope.
        :param str reference: Reference.
        """
        if isinstance(reference, Object):
            super().__init__(*CorePath._from_server_object(reference))
        elif isinstance(reference, str):
            super().__init__(scope, reference)
        elif reference is None:
            super().__init__(scope, '')
        else:
            message = 'Path validation failed: ensure the path exists and is correctly formatted.'
            logger.error(message)
            raise ValueError(message)

    @staticmethod
    def _from_server_object(reference):
        """
        Parse Path from Server Object.

        :param object reference: Path.
        :returns: Base Path and Relative Path
        :rtype: tuple(str)
        """
        classname = reference.__dict__.get('_classname', None)

        href = None
        if classname == 'ResourceInfo':
            href = reference.href
        elif classname == 'SnapshotResp':
            href = f'{reference.url}{reference.path}'
        else:
            raise ValueError('Could not determine server object: {classname}')

        href = unquote(href)
        match = re.search('^/?(ServicesPortal|admin)/webdav', href)
        start, end = match.span()
        return (href[start: end], href[end + 1:])

    @property
    def absolute(self):
        reference = self.reference.as_posix()
        previous_versions = 'PreviousVersions/'
        if previous_versions in reference:
            index = reference.index(previous_versions) + len(previous_versions)
            return f'{self.scope.as_posix()}/{quote(reference[:index]) + reference[index:]}'
        return super().absolute

    @staticmethod
    def instance(scope, entries):
        if isinstance(entries, list):
            return [CorePath(scope, e) for e in entries]
        if isinstance(entries, tuple):
            source, destination = entries
            return (CorePath(scope, source), CorePath(scope, destination))
        return CorePath(scope, entries)


class SrcDstParam(Object):

    __instance = None

    @staticmethod
    def instance(src, dest=None):
        SrcDstParam(src, dest)
        return SrcDstParam.__instance

    def __init__(self, src, dest=None):
        super().__init__()
        self._classname = self.__class__.__name__
        self.src = src
        self.dest = dest
        SrcDstParam.__instance = self  # pylint: disable=unused-private-member


class ResourceActionCursor(Object):

    def __init__(self):
        super().__init__()
        self._classname = self.__class__.__name__


class ActionResourcesParam(Object):

    __instance = None

    @staticmethod
    def instance():
        ActionResourcesParam()
        return ActionResourcesParam.__instance

    def __init__(self):
        super().__init__()
        self._classname = self.__class__.__name__
        self.urls = []
        self.startFrom = None
        ActionResourcesParam.__instance = self  # pylint: disable=unused-private-member

    def add(self, param):
        self.urls.append(param)

    def start_from(self, cursor):
        self.startFrom = cursor


class CreateShareParam(Object):

    __instance = None

    @staticmethod
    def instance(path, access, expire_on):
        CreateShareParam(path, access, expire_on)
        return CreateShareParam.__instance

    def __init__(self, path, access, expire_on):
        super().__init__()
        self._classname = self.__class__.__name__
        self.url = path
        self.share = Object()
        self.share._classname = 'ShareConfig'
        self.share.accessMode = access
        self.share.protectionLevel = 'publicLink'
        self.share.expiration = expire_on
        self.share.invitee = Object()
        self.share.invitee._classname = 'Collaborator'
        self.share.invitee.type = 'external'
        CreateShareParam.__instance = self  # pylint: disable=unused-private-member


class FetchResourcesParam(Object):

    def __init__(self):
        super().__init__()
        self._classname = 'FetchResourcesParam'
        self.start = 0
        self.limit = 100

    def increment(self):
        self.start = self.start + self.limit


class FetchResourcesParamBuilder:

    def __init__(self):
        self.param = FetchResourcesParam()

    def root(self, root):
        self.param.root = root  # pylint: disable=attribute-defined-outside-init
        return self

    def depth(self, depth):
        self.param.depth = depth  # pylint: disable=attribute-defined-outside-init
        return self

    def searchCriteria(self, criteria):
        self.param.searchCriteria = criteria  # pylint: disable=attribute-defined-outside-init
        return self

    def include_deleted(self):
        self.param.includeDeleted = True  # pylint: disable=attribute-defined-outside-init
        return self

    def limit(self, limit):
        self.param.limit = limit
        return self

    def build(self):
        return self.param


class FetchResourcesResponse(DefaultResponse):

    def __init__(self, response):
        accept_error(response.errorType)
        super().__init__(response)

    @property
    def objects(self):
        return self._response.items


@contextmanager
def fetch_resources(path, depth, include_deleted, search_criteria, limit):
    """
    Create List Directory Parameter.
    """
    depth = depth if depth is not None else 1
    builder = FetchResourcesParamBuilder().root(path.absolute_encode).depth(depth)
    if include_deleted:
        builder.include_deleted()
    if search_criteria:
        builder.searchCriteria(search_criteria)
    if limit:
        builder.limit(limit)
    logger.info('Listing directory: %s', path.reference.as_posix())
    yield builder.build()


@contextmanager
def versions(path):
    logger.info('Listing versions: %s', path.reference.as_posix())
    yield


@contextmanager
def makedir(path):
    param = Object()
    param.name = path.name
    param.parentPath = path.parent.absolute_encode
    logger.info('Creating directory: %s', path.reference.as_posix())
    yield param
    logger.info('Directory created: %s', path.reference.as_posix())


@contextmanager
def rename(path, name):
    param = ActionResourcesParam.instance()
    param.add(SrcDstParam.instance(src=path.absolute_encode, dest=path.parent.join(name).absolute_encode))
    logger.info('Renaming item: %s to: %s', path.reference.as_posix(), name)
    yield param


def _delete_or_recover(paths, *, message=None):
    param = ActionResourcesParam.instance()
    paths = [paths] if not isinstance(paths, list) else paths
    for path in paths:
        param.add(SrcDstParam.instance(src=path.absolute_encode))
        if message:
            logger.info('%s: %s', message, path.reference.as_posix())
    yield param


@contextmanager
def delete(*paths):
    return _delete_or_recover(list(paths), message='Deleting item')


@contextmanager
def recover(*paths):
    return _delete_or_recover(list(paths), message='Recovering item')


def _add_cursor_and_resolver(param, cursor, resolver):
    if cursor:
        param.startFrom = cursor

    if cursor and resolver:
        if resolver.all:
            param.startFrom.fileMoveConflictResolutaion = [resolver.build()]
        else:
            param.startFrom.skipHandler = resolver.handler


def _copy_or_move(paths, destination, resolver, cursor):
    param = ActionResourcesParam.instance()
    _add_cursor_and_resolver(param, cursor, resolver)
    for path in paths:
        src, dest = path, destination
        if isinstance(path, tuple):
            src, dest = path
        else:
            if not dest.reference.name:
                raise ValueError(f'Error: No destination specified for: {src}')
            dest = dest.join(src.name)
        param.add(SrcDstParam.instance(src=src.absolute_encode, dest=dest.absolute_encode))
    yield param


@contextmanager
def copy(*paths, destination=None, resolver=None, cursor=None):
    return _copy_or_move(list(paths), destination, resolver, cursor)


@contextmanager
def move(*paths, destination=None, resolver=None, cursor=None):
    return _copy_or_move(list(paths), destination, resolver, cursor)


@contextmanager
def public_link(path, access, expire_in):
    access = {'RO': 'ReadOnly', 'RW': 'ReadWrite', 'PO': 'PreviewOnly'}.get(access)
    expire_on = DateTimeUtils.get_expiration_date(expire_in).strftime('%Y-%m-%d')
    logger.info('Creating Public Link for: %s. Access: %s. Expires: %s', path.reference.as_posix(), access, expire_on)
    param = CreateShareParam.instance(path=path.absolute_encode, access=access, expire_on=expire_on)
    yield param


@contextmanager
def handle(path):
    logger.info('Getting file handle: %s', path.reference)
    yield path.reference


def destination_prerequisite_conditions(destination, name):
    if any(c in name for c in ['\\', '/', ':', '?', '&', '<', '>', '"', '|']):
        raise NameSyntaxError(destination.join(name).reference.as_posix())


def ensure_directory(present, resource, directory, suppress_error):
    if (not present or not resource.isFolder) and not suppress_error:
        raise NotADirectory(directory.reference.as_posix())


def ensure_writeable(resource, directory):
    if resource.scope == ResourceScope.Root:
        raise RestrictedRoot()
    if resource.scope not in [ResourceScope.Personal, ResourceScope.Project, ResourceScope.InsideCloudFolder]:
        raise UnwriteableScope(directory.reference.as_posix(), resource.scope)
    if resource.permission != FileAccessMode.RW:
        raise PermissionDenied(directory.reference.as_posix(), ResourceAction.Write)


@contextmanager
def upload(core, name, destination, size, fd):
    fd, size = common.encode_stream(fd, size)
    param = dict(
        name=name,
        Filename=name,
        fullpath=core.io.builder(CorePath(destination.reference, name).absolute_encode),
        fileSize=size,
        file=fd
    )
    logger.info('Uploading: %s to: %s', name, destination.reference)
    yield param


def validate_transfer_success(response, path):
    if response.rc:
        logger.error('Upload of file: "%s" failed.', path)
        if response.msg == UploadError.UserQuotaViolation:
            raise OutOfQuota('User', path)
        if response.msg == UploadError.PortalQuotaViolation:
            raise OutOfQuota('Team Portal', path)
        if response.msg == UploadError.FolderQuotaViolation:
            raise OutOfQuota('Cloud drive folder', path)
        if response.msg == UploadError.RejectedByPolicy:
            raise RejectedByPolicy(path)
        if response.msg == UploadError.WindowsACL:
            raise WindowsACLError(path)
        if response.msg.startswith(UploadError.NoStorageBucket):
            raise NoStorageBucket(path)
        raise UploadException(f'Upload failed. Reason: {response.msg}', path)
    if not response.rc and response.msg == 'OK':
        logger.info('Upload successful. Saved to: %s', path)


@contextmanager
def handle_many(directory, objects):
    param = Object()
    param.paths = ['/'.join([directory.absolute, filename]) for filename in objects]
    param.snapshot = None
    param.password = None
    param.portalName = None
    param.showDeleted = False
    logger.info('Getting directory handle: %s', directory.reference)
    yield param


@contextmanager
def share_info(path):
    logger.info('Getting Share info: %s', path.reference.as_posix())
    yield path.absolute_encode


def _is_sharing_on_user_behalf(path):
    return re.match(r'^Users/[^/]+(?=/)', path.reference.as_posix()) is not None


def _is_subfolder(path):
    depth = len(path.reference.parts)
    sharing_on_user_behalf = _is_sharing_on_user_behalf(path)
    return (sharing_on_user_behalf and depth > 3) or (not sharing_on_user_behalf and depth > 1)


@contextmanager
def share(path, as_project, allow_reshare, allow_sync, shares=None):
    if _is_subfolder(path):
        as_project = False  # Can't be a team project
        allow_sync = False  # Can't allow sync of a sub-folder

    param = Object()
    param._classname = 'ShareResourceParam'  # pylint: disable=protected-access
    param.url = path.absolute_encode
    param.teamProject = as_project
    param.allowReshare = allow_reshare
    param.shouldSync = allow_sync
    param.shares = []
    if shares is not None:
        for recipient in shares:
            settings = Object()
            settings._classname = 'ShareConfig'  # pylint: disable=protected-access
            settings.accessMode = recipient.accessMode
            settings.expiration = recipient.expiration
            settings.protectionLevel = recipient.protectionLevel
            settings.invitee = recipient.invitee
            param.shares.append(settings)

    yield param


@contextmanager
def unshare(resource_info, path):
    with share(path, True, True, True) as param:
        logger.info("Unsharing %s: %s", ('folder' if resource_info.isFolder else 'file'), path.reference.as_posix())
        yield param


def add_share_recipient(param, recipient):
    settings = Object()
    settings._classname = 'ShareConfig'  # pylint: disable=protected-access
    settings.accessMode = recipient.access
    if not recipient.access == FileAccessMode.NA:
        settings.expiration = recipient.expiration_date
    invitee = None
    if recipient.type == CollaboratorType.EXT:
        invitee = Object()
        invitee._classname = 'Collaborator'  # pylint: disable=protected-access
        invitee.type = recipient.type
        invitee.email = recipient.account
        if recipient.two_factor:
            settings.protectionLevel = ProtectionLevel.Email
        else:
            settings.protectionLevel = ProtectionLevel.Public
    else:
        invitee = recipient.collaborator
    settings.invitee = invitee
    param.shares.append(settings)


def find_recipients_to_remove(param, path, current_accounts, accounts):
    current_accounts = obtain_current_accounts(param)
    accounts_to_keep, accounts_to_remove = [], []
    if current_accounts:
        for idx, current_account in enumerate(current_accounts):
            if current_account not in accounts:
                accounts_to_keep.append(param.shares[idx])
            else:
                logger.debug('Found recipient to remove: %s', str(current_account))
                accounts_to_remove.append(current_account)
        if not accounts_to_remove:
            logger.info('Share recipients not found: %s', [str(member) for member in accounts])
        else:
            logger.info('Removing share recipients: %s', [str(account) for account in accounts_to_remove])
    else:
        logger.info('Share has no recipients. %s', path.reference.as_posix())
    return accounts_to_keep, accounts_to_remove


def find_recipients_to_add(path, param, current_accounts, valid_recipients):
    current_accounts = obtain_current_accounts(param)
    accounts_to_add = []
    if valid_recipients:
        for recipient in valid_recipients:
            if recipient.account not in current_accounts:
                logger.debug('New share recipient: %s', str(recipient.account))
                accounts_to_add.append(recipient)
            else:
                logger.debug('Share recipient already exists: %s', str(recipient.account))
        if accounts_to_add:
            logger.info('Adding share recipients: %s', [str(recipient.account) for recipient in accounts_to_add])
        else:
            logger.warning('Could not find new share recipients: %s', path.reference.as_posix())
        return accounts_to_add
    logger.warning('Could not find valid share recipients: %s', path.reference.as_posix())
    return accounts_to_add


@contextmanager
def search_collaboration_member(account, cloud_folder_uid):
    param = Object()
    if account.account_type == PortalAccountType.User:
        param.searchType = SearchType.Users
    elif account.account_type == PortalAccountType.Group:
        param.searchType = SearchType.Groups
    else:
        raise ValueError(f'Invalid account type: {account.account_type}')

    param.searchTerm = account.name
    param.resourceUid = cloud_folder_uid
    param.countLimit = 1
    logger.info('Searching for member: %s', str(account))
    yield param


def consume_search_collaboration_response(response, account):
    if not response.objects:
        logger.warning('Could not find results that match your search criteria. %s', {'name': account.name})
    elif len(response.objects) > 1:
        logger.warning('Search returned multiple results. Skipping...')
    else:
        collaborator = response.objects.pop()
        if account.is_local:
            if account.account_type == PortalAccountType.User and collaborator.type == CollaboratorType.LU:
                logger.debug('Found local user account. %s', {'name': account.name})
                return collaborator
            if account.account_type == PortalAccountType.Group and collaborator.type == CollaboratorType.LG:
                logger.debug('Found local group account. %s', {'name': account.name})
                return collaborator
            logger.warning('Expected to find local %s but found domain %s', account.account_type, account.account_type)
        else:
            if account.account_type == PortalAccountType.User and collaborator.type == CollaboratorType.DU:
                logger.debug('Found domain user account. %s', {'name': account.name, 'domain': account.directory})
                return collaborator
            if account.account_type == PortalAccountType.Group and collaborator.type == CollaboratorType.DG:
                logger.debug('Found domain group account. %s', {'name': account.name, 'domain': account.directory})
                return collaborator
            logger.warning('Expected to find domain %s but found local %s', account.account_type, account.account_type)
    return None


def valid_recipient(recipient):
    is_valid = True
    if not recipient.account:
        logger.warning('No account information found. Skipping. %s', {'account': recipient.account})
        is_valid = False
    if not isinstance(recipient.account, (str, UserAccount, GroupAccount)):
        logger.warning('Invalid recipient type. Skipping. %s', {'type': type(recipient.account)})
        is_valid = False
    if recipient.access not in [v for k, v in FileAccessMode.__dict__.items() if not k.startswith('_')]:
        logger.warning('Invalid file access mode. Skipping. %s', {'account': str(recipient.account), 'access': recipient.access})
        is_valid = False
    if recipient.type == CollaboratorType.EXT and not isinstance(recipient.account, str):
        logger.warning('Invalid external recipient type. Skipping. %s', {'expected': 'str', 'actual': type(recipient.account)})
        is_valid = False
    if recipient.type in [CollaboratorType.LU, CollaboratorType.LG] and not recipient.account.is_local:
        logger.warning('Expected local account. Received a domain account. Skipping. %s', {'account': str(recipient.account)})
        is_valid = False
    if recipient.type in [CollaboratorType.DU, CollaboratorType.DG] and recipient.account.is_local:
        logger.warning('Expected domain account. Received a local account. Skipping. %s', {'account': str(recipient.account)})
        is_valid = False
    return is_valid


def obtain_current_accounts(param):
    current_accounts = []
    for collaborator in param.shares:
        if collaborator.invitee.type == CollaboratorType.EXT:
            current_accounts.append(collaborator.invitee.email)
        else:
            current_accounts.append(PortalAccount.from_collaborator(collaborator.invitee))
    return current_accounts


file_access_errors = {
    "Conflict": FileConflict,
    "PermissionDenied": PermissionDenied,
    "DestinationNotExists": PathValidationError,
    "FileWithTheSameNameExist": ResourceExistsError,
    "InvalidName": NameSyntaxError,
    "ReservedName": ReservedNameError
}


def await_or_future(core, ref, wait):
    """
    Wait for task completion, or return an awaitable task object.

    :param str ref: Task reference
    :param bool wait: ``True`` to wait for task completion, ``False`` to return an awaitable task object
    """
    if wait:
        task = core.tasks.wait(ref)
        accept_error(task.error_type, **error_metadata(task))
        return task
    return core.tasks.awaitable_task(ref)


async def a_await_or_future(ctera, ref, wait):
    """
    Wait for task completion, or return an awaitable task object.

    :param str ref: Task reference
    :param bool wait: ``True`` to wait for task completion, ``False`` to return an awaitable task object
    """
    if wait:
        task = await ctera.tasks.wait(ref)
        accept_error(task.error_type, **error_metadata(task))
        return task
    return ctera.tasks.awaitable_task(ref)


def error_metadata(task):
    metadata = dict(
        action=task.name
    )
    if task.name in [ResourceAction.Copy, ResourceAction.Move]:
        metadata.update(dict(cursor=task.cursor))
        if task.cursor.destResource:
            metadata.update(dict(name=task.cursor.destResource.name))
    return metadata


def accept_error(error_type, **kwargs):
    """
    Check if response contains an error.
    """
    if error_type not in [None, 'OK']:
        exception_classname = file_access_errors.get(error_type, RemoteStorageError)
        try:
            raise exception_classname(**kwargs)
        except ResourceExistsError as error:
            logger.info('Resource already exists: a file or folder with this name already exists.')
            raise error
        except PathValidationError as error:
            logger.error('Path validation failed: the specified destination path does not exist.')
            raise error
        except NameSyntaxError as error:
            logger.error('Invalid name: the name contains characters that are not allowed.')
            raise error
        except ReservedNameError as error:
            logger.error('Reserved name error: the name is reserved and cannot be used.')
            raise error
        except PermissionDenied as error:
            logger.error('Permission denied: Inappropriate permissions to access this resource.')
            raise error
        except FileConflict as error:
            logger.error('Conflict: a file with the same name already exists: %s', error.name)
            raise error
        except RemoteStorageError as error:
            logger.error('An error occurred while performing operation.')
            raise error



from . import (  # noqa: E402, F401
    auth,
    backup,
    common,
    direct,
    io,
    notifications,
    session,
    transport
)

from .base import CTERAException


class ObjectNotFoundException(CTERAException):
    """
    Object not found.

    :param str urn: Resource
    """

    def __init__(self, urn):
        super().__init__(f'Object not found: {urn}')


class InputError(ValueError):
    """
    Input Error

    :param str message: Error message
    :param object expression: Error expression
    :param object options: Options
    """

    def __init__(self, message, expression=None, options=None):
        super().__init__(self, message)
        self.expression = expression
        self.options = options


class UserConsentError(CTERAException):
    """Console"""



from http import HTTPStatus
from .base import CTERAException


class HTTPError(CTERAException):
    """
    HTTP Error

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, status, error):
        super().__init__(error.request.url)
        self.code = status.value
        self.name = status.name
        self.error = error


class BadRequest(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.BAD_REQUEST, error)


class Unauthorized(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.UNAUTHORIZED, error)


class Forbidden(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.FORBIDDEN, error)


class NotFound(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.NOT_FOUND, error)


class NotAllowed(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.METHOD_NOT_ALLOWED, error)


class PreConditionFailed(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.PRECONDITION_FAILED, error)


class Unprocessable(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.UNPROCESSABLE_ENTITY, error)


class InternalServerError(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.INTERNAL_SERVER_ERROR, error)


class BadGateway(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.BAD_GATEWAY, error)


class ServiceUnavailable(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.SERVICE_UNAVAILABLE, error)


class GatewayTimeout(HTTPError):

    def __init__(self, error):
        super().__init__(HTTPStatus.GATEWAY_TIMEOUT, error)


class TLSError(CTERAException):
    """
    TLS Error

    :ivar str host: Host
    :ivar int port: Port
    """

    def __init__(self, host, port):
        super().__init__(f"TLS handshake to '{host}:{port}' failed.")
        self.host = host
        self.port = port



import logging
import asyncio

from ..lib.retries import execute_with_retries
from .types import Metadata, Block
from .crypto import decrypt_key, decrypt_block
from .decompressor import decompress
from ..exceptions.transport import BadRequest, Unauthorized, Unprocessable, InternalServerError, HTTPError
from ..exceptions.direct import (
    AuthorizationError, BlockListConnectionError, BlockListTimeout, BlockValidationException, BlocksNotFoundError,
    DecompressBlockError, DecryptBlockError, DecryptKeyError, DirectIOError, DownloadConnectionError,
    DownloadError, DownloadTimeout, InvalidRequest, ObjectNotFoundError, UnsupportedStorageError
)


logger = logging.getLogger('cterasdk.direct')


@execute_with_retries(retries=3, backoff=1, max_backoff=10)
async def get_object(client, file_id, chunk):
    """
    Get Object from a Signed URL.

    :param int file_id: File ID.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Object
    :rtype: bytes
    """
    message = (
        f"Downloading block #{chunk.number} "
        f"(offset={chunk.offset}, length={chunk.length})"
    )

    if file_id:
        message += f" for file ID {file_id}"

    error_message, exception = None, None

    logger.debug(message)
    try:
        response = await client.get(chunk.url)
        return await response.read()
    except ConnectionError:
        error_message = 'connection'
        exception = DownloadConnectionError(file_id, chunk)
    except asyncio.TimeoutError:
        error_message = 'timeout'
        exception = DownloadTimeout(file_id, chunk)
    except IOError as error:
        error_message = 'io'
        exception = DownloadError(error, file_id, chunk)
    except HTTPError as error:
        error_message = 'unknown'
        exception = DownloadError(error.error, file_id, chunk)

    error_messages = {
        "connection": "Connection error",
        "timeout": "Timed out",
        "io": "I/O error",
        "unknown": "Unknown error"
    }

    message = (
        f"Failed to download block #{chunk.number} "
        f"(offset={chunk.offset}, length={chunk.length})"
    )
    if file_id:
        message = message + f" for file ID {file_id}"

    message = message + f": {error_messages.get(error_message, 'Unknown error')}."
    logger.error(message)
    raise exception


async def decrypt_object(file_id, encrypted_object, encryption_key, chunk):
    """
    Decrypt Encrypted Object.

    :param bytes encrypted_object: Encrypted object.
    :param bytes encryption_key: Encryption key.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Decrypted Object.
    :rtype: bytes
    """
    try:
        return decrypt_block(encrypted_object, encryption_key)
    except DirectIOError:
        logger.error('Failed to decrypt block.')
        raise DecryptBlockError(file_id, chunk)


async def decompress_object(file_id, compressed_object, chunk):
    """
    Decompress Object.

    :param bytes compressed_object: Compressed object.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Decompressed Object.
    :rtype: bytes
    """
    try:
        decompressed_object = decompress(compressed_object)
        if chunk.length != len(decompressed_object):
            logger.error('Expected block length does not match decrypted and decompressed block length.')
            raise BlockValidationException(file_id, chunk)
        return decompressed_object
    except DirectIOError:
        logger.error('Failed to decompress block.')
        raise DecompressBlockError(file_id, chunk)


async def process_chunk(client, file_id, chunk, encryption_key, semaphore):
    """
    Process a Chunk.

    :param cterasdk.clients.clients.AsyncClient client: Asynchronous HTTP Client.
    :param int file_id: File ID.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :param str encryption_key: Encryption key.
    :param asyncio.Semaphore semaphore: Semaphore.

    :returns: Block
    :rtype: cterasdk.direct.types.Block
    """
    async def process(client, chunk, encryption_key):
        message = (
            f"Processing block #{chunk.number} "
            f"(offset={chunk.offset}, length={chunk.length})"
        )
        if file_id:
            message = message + f" for file ID {file_id}"
        logger.debug(message)
        encrypted_object = await get_object(client, file_id, chunk)
        decrypted_object = await decrypt_object(file_id, encrypted_object, encryption_key, chunk)
        decompressed_object = await decompress_object(file_id, decrypted_object, chunk)
        return Block(file_id, chunk.number, chunk.offset, decompressed_object, chunk.length)

    if semaphore is not None:
        async with semaphore:
            return await process(client, chunk, encryption_key)
    return await process(client, chunk, encryption_key)


async def process_chunks(client, file_id, chunks, encryption_key, semaphore=None):
    """
    Process Chunks Asynchronously.

    :param cterasdk.clients.clients.AsyncClient client: Asynchronous HTTP Client.
    :param int file_id: File ID.
    :param list[cterasdk.direct.types.Chunk] chunks: Chunk.
    :param str encryption_key: Encryption key.
    :param asyncio.Semaphore,optional semaphore: Semaphore.
    :returns: List of futures.
    :rtype: list[asyncio.Task]
    """
    message = [f"Processing {len(chunks)} blocks"]
    if file_id:
        message.append(f"for file ID {file_id}")
    if semaphore:
        message.append(f"using up to {semaphore._value} workers")  # pylint: disable=protected-access
    logger.debug(' '.join(message))
    futures = []
    for chunk in chunks:
        futures.append(asyncio.create_task(process_chunk(client, file_id, chunk, encryption_key, semaphore)))
    return futures


def decrypt_encryption_key(file_id, wrapped_key, secret_access_key):
    """
    Decrypt Encryption Key.

    :param int file_id: File ID.
    :param str wrapped_key: Encryption Key.
    :param str secret_access_key: Secret Access Key.
    :returns: Decrypted Encryption Key.
    :rtype: bytes
    """
    try:
        return decrypt_key(wrapped_key, secret_access_key)
    except DirectIOError:
        logger.error('Failed to decrypt secret key.')
        raise DecryptKeyError(file_id)


@execute_with_retries(retries=3, backoff=1, max_backoff=10)
async def get_chunks(api, bearer, file_id):
    """
    Get Chunks.

    :param cterasdk.clients.clients.AsyncJSON api: Asynchronous JSON Client.
    :param str bearer: Bearer Token.
    :param int file_id: File ID.
    :returns: Wrapped key and file chunks.
    :rtype: cterasdk.direct.types.Metadata
    """
    logger.debug('Listing blocks for file ID: %s', file_id)
    try:
        response = await api.get(f'{file_id}', headers={'Authorization': bearer})
        if not response.chunks:
            logger.error('Could not find blocks for file ID: %s.', file_id)
            raise BlocksNotFoundError(file_id)
        return Metadata(file_id, response)
    except BadRequest as error:
        raise ObjectNotFoundError(file_id) from error
    except Unauthorized as error:
        raise AuthorizationError(file_id) from error
    except Unprocessable as error:
        raise UnsupportedStorageError(file_id) from error
    except InternalServerError as error:
        raise InvalidRequest(file_id) from error
    except ConnectionError:
        logger.error('Failed to list blocks for file ID: %s due to a connection error.', file_id)
        raise BlockListConnectionError(file_id)
    except asyncio.TimeoutError:
        logger.error('Timed out while listing blocks for file ID: %s.', file_id)
        raise BlockListTimeout(file_id)



import re
import json
import logging
from datetime import datetime
from collections.abc import MutableMapping


logger = logging.getLogger('cterasdk.common')


class ObjectEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        if o.get('__dict__', None):
            return o.__dict__
        return super().default(o)


class Object(MutableMapping):  # pylint: disable=too-many-instance-attributes

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __str__(self):
        return json.dumps(self, cls=ObjectEncoder, indent=5)

    def __repr__(self):
        return str(self)


class Device(Object):

    def __init__(self, uid, version, firmware):
        super().__init__()
        self.namespace = 'http://www.w3.org/2001/XMLSchema-instance'
        self.location = '../../db/resources/db.xsd'
        self.id = uid
        self.version = version
        self.firmware = firmware


def delete_attrs(obj, paths):
    """
    Delete attributes

    :param cterasdk.common.object.Object object: The object
    :param list[str] paths: List of attributes to remove

    :returns: The modified object
    :rtype: cterasdk.common.object.Object
    """
    for path in paths:
        delete_attr(obj, path)


def delete_attr(obj, path):
    """
    Delete attribute

    :param cterasdk.common.object.Object object: The object
    :param str path: Attribute path

    :returns: The modified object
    :rtype: cterasdk.common.object.Object
    """
    parts = re.findall('[^/]+', path)
    parent = find_attr(obj, parts[:-1])
    remove_attr(parent, parts[-1])

    if len(parts) > 1 and isinstance(parent, Object) and not parent.__dict__:
        grandparent = find_attr(obj, parts[:-2])
        setattr(grandparent, parts[-2], None)


def find_attr(obj, path):
    """
    Find attribute

    :param cterasdk.common.object.Object object: The object
    :param str path: A string or an array of the attribute path

    :returns: The attribute, or ``None`` if not found
    """
    parts = re.findall('[^/]+', path) if isinstance(path, str) else path

    attr = obj
    for part in parts:
        attr = get_attr(attr, part)
        if attr is None:
            logger.warning('Could not find attribute. %s', {'path': f'/{"/".join(parts)}'})
            return attr
    return attr


def get_attr(obj, attr):
    """
    Get attribute

    :param cterasdk.common.object.Object object: The object
    :param str attr: The name of the attribute to retrieve

    :returns: The attribute, or ``None`` if not found
    """
    if isinstance(obj, list):
        try:
            attr = int(attr)
            return obj[attr]
        except ValueError:
            logger.warning('Could not find attribute.')
            return None

    return getattr(obj, attr, None)


def remove_attr(obj, attr):
    """
    Remove attribute

    :param cterasdk.common.object.Object object: The object
    :param str attr: The name of the attribute to remove
    """
    if isinstance(obj, list):
        remove_array_element(obj, attr)
    else:
        try:
            delattr(obj, attr)
        except AttributeError:
            logger.warning('Failed to remove attribute. Attribute not found. %s', {'attr': attr})


def remove_array_element(array, attr):

    try:
        attr = int(attr)
        if attr <= len(array) - 1:
            array.pop(attr)
        else:
            logger.warning('Could not remove array item. Index out of range. %s', {'index': attr})
    except ValueError:
        pass

    if remove_array_element_by_key(array, '_uuid', attr):
        return

    remove_array_element_by_key(array, 'name', attr)


def remove_array_element_by_key(array, key, value):
    for index, element in enumerate(array):
        element_value = getattr(element, key, None)
        if element_value == value:
            return array.pop(index)
    return None



from .base import CTERAException


class NotFound(CTERAException):
    """ Not found exception """


class AttachEncrypted(CTERAException):
    """ Attach Encrypted exception """

    def __init__(self, encryptionMode, encryptedFolderKey, passPhraseSalt):
        super().__init__()
        self.encryptionMode = encryptionMode
        self.encryptedFolderKey = encryptedFolderKey
        self.passPhraseSalt = passPhraseSalt


class IncorrectPassphrase(CTERAException):
    """ Incorrect Passphrase exception """

    def __init__(self):
        super().__init__('Incorrect passphrase')


class ClocksOutOfSync(CTERAException):
    """ Clocks Out of Sync exception """

    def __init__(self):
        super().__init__('Clocks are out of sync')



from collections import namedtuple
from . import enum
from ..common import Object
from ..exceptions import InputError


TCPService = namedtuple('TCPService', ('host', 'port'))
TCPService.__doc__ = 'Tuple holding the host and port to connect over TCP'
TCPService.host.__doc__ = 'The ip address, hostname or fully qualified domain name of the host'
TCPService.port.__doc__ = 'The port number'


TCPConnectResult = namedtuple('TCPConnectResult', ('host', 'port', 'is_open'))
TCPConnectResult.__doc__ = 'Tuple holding the host and port to connect over TCP'
TCPConnectResult.host.__doc__ = 'The ip address, hostname or fully qualified domain name of the host'
TCPConnectResult.port.__doc__ = 'The port number'
TCPConnectResult.is_open.__doc__ = 'Boolean, indicating whether a TCP connection can be successfully ' \
                                   'established to the target host over the specified port'


class UserGroupEntry():
    """
    User or Group Entry

    :ivar cterasdk.edge.enum.PrincipalType principal_type: Principal type of the ACL
    :ivar str name: The name of the user or group
    """

    _valid_principal_types = list({k: v for k, v in enum.PrincipalType.__dict__.items() if not k.startswith('_')}.values())

    def __init__(self, principal_type, name):
        UserGroupEntry._validate_principal_type(principal_type)
        self._principal_type = principal_type
        self.name = name

    @property
    def principal_type(self):
        return self._principal_type

    @principal_type.setter
    def principal_type(self, principal_type):
        UserGroupEntry._validate_principal_type(principal_type)
        self._principal_type = principal_type

    def to_server_object(self):
        user_group_obj = Object()
        if self.principal_type == enum.PrincipalType.LU:
            user_group_obj._classname = enum.PrincipalType.LU  # pylint: disable=protected-access
            user_group_obj.ref = "#config#auth#users#" + self.name
        elif self.principal_type == enum.PrincipalType.LG:
            user_group_obj._classname = enum.PrincipalType.LG  # pylint: disable=protected-access
            user_group_obj.ref = "#config#auth#groups#" + self.name
        elif self.principal_type == enum.PrincipalType.DU:
            user_group_obj._classname = enum.PrincipalType.DU  # pylint: disable=protected-access
            user_group_obj.name = self.name
        elif self.principal_type == enum.PrincipalType.DG:
            user_group_obj._classname = enum.PrincipalType.DG  # pylint: disable=protected-access
            user_group_obj.name = self.name
        return user_group_obj

    @staticmethod
    def from_server_object(server_object):
        principal_type = server_object._classname  # pylint: disable=protected-access
        if principal_type in [enum.PrincipalType.LU, enum.PrincipalType.LG]:
            name = server_object.ref
            name = name[name.rfind('#') + 1:]
        else:
            name = server_object.name
        return UserGroupEntry(principal_type, name)

    @staticmethod
    def _validate_principal_type(principal_type):
        if principal_type not in UserGroupEntry._valid_principal_types:
            raise InputError('Invalid principal type', principal_type, UserGroupEntry._valid_principal_types)

    def __str__(self):
        return ('\\' + self.name) if self.principal_type in [enum.PrincipalType.LG, enum.PrincipalType.LU] else self.name


class AccessControlEntryValidator:

    _valid_permissions = list({k: v for k, v in enum.FileAccessMode.__dict__.items() if not k.startswith('_')}.values())

    @staticmethod
    def validate_permission(permission):
        if permission not in AccessControlEntryValidator._valid_permissions:
            raise InputError('Invalid permissions', permission, AccessControlEntryValidator._valid_permissions)


class NFSv3AccessControlEntry():
    """
    NFS v3 export access control entry
    :ivar str address: IP address, hostname or fully qualified domain name of client machine
    :ivar str netmask: Subnet mask
    :ivar cterasdk.edge.enum.FileAccessMode perm: File access permission
    :ivar bool insecure: Allow insecure NFS connections
    """

    def __init__(self, address, netmask, perm, insecure=False):
        AccessControlEntryValidator.validate_permission(perm)
        self._address = address
        self._netmask = netmask
        self._perm = perm
        self._noRootSquash = False
        self._insecure = insecure

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, address):
        self._address = address

    @property
    def netmask(self):
        return self._netmask

    @netmask.setter
    def netmask(self, netmask):
        self._netmask = netmask

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, perm):
        AccessControlEntryValidator.validate_permission(perm)
        self._perm = perm

    @property
    def insecure(self):
        return self._insecure

    @insecure.setter
    def insecure(self, insecure):
        self._insecure = insecure

    @staticmethod
    def from_server_object(server_object):
        return NFSv3AccessControlEntry(
            server_object.address,
            server_object.netmask,
            server_object.accessLevel,
            getattr(server_object, 'insecure', False)  # Use getattr for backward compatibility
        )

    def to_server_object(self):
        param = Object()
        param.address = self._address
        param.netmask = self._netmask
        param.accessLevel = self._perm
        param.noRootSquash = self._noRootSquash
        param.insecure = self._insecure
        return param

    def __str__(self):
        return str(
            dict(
                address=self._address,
                netmask=self._netmask,
                permission=self._perm,
                insecure=self._insecure
            )
        )


class RemoveNFSv3AccessControlEntry():
    """
    Object holding address and netmasak for NFS v3 export access control entry
    :ivar str address: IP address, hostname or fully qualified domain name of client machine
    :ivar str netmask: Subnet mask
    """

    def __init__(self, address, netmask):
        self._address = address
        self._netmask = netmask

    @property
    def address(self):
        return self._address

    @property
    def netmask(self):
        return self._netmask


class ShareAccessControlEntry():
    """
    Share access control entry for filer shares

    :ivar cterasdk.edge.enum.PrincipalType principal_type: Principal type of the ACL
    :ivar str name: The name of the user or group
    :ivar cterasdk.edge.enum.FileAccessMode perm: The file access permission
    """

    _valid_permissions = list({k: v for k, v in enum.FileAccessMode.__dict__.items() if not k.startswith('_')}.values())

    def __init__(self, principal_type, name, perm):
        AccessControlEntryValidator.validate_permission(perm)
        self._user_group_entry = UserGroupEntry(principal_type, name)
        self._perm = perm

    @property
    def principal_type(self):
        return self._user_group_entry.principal_type

    @principal_type.setter
    def principal_type(self, principal_type):
        self._user_group_entry.principal_type = principal_type

    @property
    def name(self):
        return self._user_group_entry.name

    @name.setter
    def name(self, name):
        self._user_group_entry.name = name

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, perm):
        AccessControlEntryValidator.validate_permission(perm)
        self._perm = perm

    def to_server_object(self):
        ace = Object()
        ace._classname = "ShareACLRule"  # pylint: disable=protected-access
        ace.principal2 = self._user_group_entry.to_server_object()
        ace.permissions = self._create_permissions_object()
        return ace

    def _create_permissions_object(self):
        permissions = Object()
        permissions._classname = "FileAccessPermissions"  # pylint: disable=protected-access
        permissions.allowedFileAccess = self.perm
        return permissions

    @staticmethod
    def from_server_object(server_object):
        user_group_entry = UserGroupEntry.from_server_object(server_object.principal2)
        perm = server_object.permissions.allowedFileAccess
        return ShareAccessControlEntry(user_group_entry.principal_type, user_group_entry.name, perm)


class RemoveShareAccessControlEntry(UserGroupEntry):
    """
    Object holding share access control principal type and name

    :ivar cterasdk.edge.enum.PrincipalType principal_type: Principal type of the ACL
    :ivar str name: The name of the user or group
    """


class HostCredentials:
    """
    Source Host Credential Object

    :ivar cterasdk.edge.enum.SourceType host_type: Host type
    :ivar str host: Fully qualified domain name, hostname or IP address
    :ivar str username: Source host account username
    :ivar str password: Source host account password
    """

    def __init__(self, host, username, password, host_type=None):
        self._host = host
        self._username = username
        self._password = password
        self._host_type = host_type

    @property
    def host_type(self):
        return self._host_type

    @property
    def host(self):
        return self._host

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @staticmethod
    def localhost():
        return HostCredentials('127.0.0.1', 'dummy', 'dummy', enum.SourceType.Edge)


class DeduplicationStatus(Object):
    """
    Edge Filer Local Deduplication Status Object

    :ivar bool enabled: ``True`` if deduplication is enabled, ``False`` otherwise
    :ivar int size: Logical Size in Bytes
    :ivar int usage: Actual Size in Bytes
    """

    def __init__(self, enabled, size, usage):
        super().__init__()
        self.enabled = enabled
        self.size = size
        self.usage = usage

        savings, dedup = 0, 0
        if self.usage < self.size:
            dedup = self.size - self.usage
            savings = 1 - self.usage / self.size

        self.dedup = dedup
        self.savings = f"{savings:.2%}"


class AntivirusUpdateSchedule(Object):
    """
    Edge Filer Antivirus Update Schedule
    """

    @staticmethod
    def daily(hour, minute):
        """
        Daily update.

        :param int hour: Hour
        :param int minute: Minute
        """
        return AntivirusUpdateSchedule(mode='daily', daily=Object(hour=hour, minute=minute))

    @staticmethod
    def weekly(day, hour, minute):
        """
        Weekly update.

        :param cterasdk.common.enum.DayOfWeek day: Day
        :param int hour: Hour
        :param int minute: Minute
        """
        return AntivirusUpdateSchedule(mode='weekly', weekly=Object(day=day, hour=hour, minute=minute))

    @staticmethod
    def monthly(day, hour, minute):
        """
        Monthly update.

        :param int day: Day
        :param int hour: Hour
        :param int minute: Minute
        """
        return AntivirusUpdateSchedule(mode='monthly', monthly=Object(day=day, hour=hour, minute=minute))


class AlertSettings(Object):  # pylint: disable=too-many-instance-attributes
    """
    Alert Settings

    :ivar bool firmware_upgrade: Enables alerting when a firmware upgrade occurs.
    :ivar bool device_startup: Enables alerting on Edge Filer shutdowns and startups.
    :ivar bool backup_success: Enables alerting upon successful backup completion.
    :ivar bool storage_volume_usage: Enables alerting for high storage volume utilization.
    :ivar int storage_volume_usage_percent: Volume usage percentage threshold that triggers an alert.
    :ivar bool overdue_backup: Enables alerting when a Cloud Backup is overdue.
    :ivar int overdue_backup_days: Number of days since last Cloud Backup after which an alert is triggered.
    :ivar bool delayed_synchronization: Enables alerting when synchronization is delayed.
    :ivar int delayed_synchronization_hours: Number of hours of synchronization delay to trigger an alert.
    :ivar bool disconnected: Enables alerting when the Edge Filer disconnects from the CTERA Portal.
    :ivar int disconnected_hours: Number of hours of disconnection duration to trigger an alert.
    """
    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(self, firmware_upgrade, device_startup, backup_success,
                 storage_volume_usage, storage_volume_usage_percent,
                 overdue_backup, overdue_backup_days,
                 delayed_synchronization, delayed_synchronization_hours,
                 disconnected, disconnected_hours):
        super().__init__()
        self.firmware_upgrade = firmware_upgrade
        self.device_startup = device_startup
        self.backup_success = backup_success
        self.storage_volume_usage = storage_volume_usage
        self.storage_volume_usage_percent = storage_volume_usage_percent
        self.overdue_backup = overdue_backup
        self.overdue_backup_days = overdue_backup_days
        self.delayed_synchronization = delayed_synchronization
        self.delayed_synchronization_hours = delayed_synchronization_hours
        self.disconnected = disconnected
        self.disconnected_hours = disconnected_hours

    def to_server_object(self):
        param = Object()
        param._classname = 'SpecificAlerts'  # pylint: disable=protected-access
        param.NotifyFirmwareUpgrade = self.firmware_upgrade
        param.NotifyDeviceStarted = self.device_startup
        param.NotifyBackupSuccess = self.backup_success
        param.VolumeFullAlert = self.storage_volume_usage
        param.VolumeFullPercent = self.storage_volume_usage_percent
        param.BackupFailAlert = self.overdue_backup
        param.BackupFailDays = self.overdue_backup_days
        param.CloudSyncFailAlert = self.delayed_synchronization
        param.CloudSyncFailHours = self.delayed_synchronization_hours
        param.CloudConnectFailAlert = self.disconnected
        param.CloudConnectFailHours = self.disconnected_hours
        return param

    @staticmethod
    def from_server_object(server_object):
        params = {
            'firmware_upgrade': server_object.NotifyFirmwareUpgrade,
            'device_startup': server_object.NotifyDeviceStarted,
            'backup_success': server_object.NotifyBackupSuccess,
            'storage_volume_usage': server_object.VolumeFullAlert,
            'storage_volume_usage_percent': server_object.VolumeFullPercent,
            'overdue_backup': server_object.BackupFailAlert,
            'overdue_backup_days': server_object.BackupFailDays,
            'delayed_synchronization': server_object.CloudSyncFailAlert,
            'delayed_synchronization_hours': server_object.CloudSyncFailHours,
            'disconnected': server_object.CloudConnectFailAlert,
            'disconnected_hours': server_object.CloudConnectFailHours,
        }
        return AlertSettings(**params)



from datetime import datetime

from ..lib import QueryIterator, DefaultResponse, Command
from ..common import Object


def create_callback_function(core, path, name=None, *, callback_response=None):
    """
    Create a query callback function

    :param cterasdk.objects.core.Portal core: Portal object
    :param cterasdk.core.query.QueryParams param: Query paramter object
    :param str,optional name: Schema method name
    :param cterasdk.lib.iterator.BaseResponse callback_response: Class to consume callback response

    :returns: Command object
    """

    def database(core, path, name, param):
        return callback_response(core.api.database(path, name, param))

    def execute(core, path, name, param):
        return callback_response(core.api.execute(path, name, param))

    return Command(execute if name else database, core, path, name or 'query')


def iterator(core, path, param=None, name=None, *, callback_response=None):
    """
    Create iterator

    :param cterasdk.objects.core.Portal core: Portal object
    :param str path: URL Path
    :param str,optional name: Schema method name
    :param cterasdk.core.query.QueryParams,optional param: Query paramter object
    :param cterasdk.lib.iterator.BaseResponse callback_response: Class to consume callback response

    :returns: Query iterator object
    """

    callback_response = callback_response if callback_response else DefaultResponse
    callback_function = create_callback_function(core, path, name, callback_response=callback_response)
    return QueryIterator(callback_function, param if param else QueryParams())


class Restriction:
    LIKE = "like"
    UNLIKE = "notLike"
    EQUALS = "eq"
    NOT_EQUALS = "ne"
    GREATER_EQUALS = "ge"
    GREATER_THAN = "gt"
    LESS_EQUALS = "le"
    LESS_THAN = "lt"


class FilterType:

    DateTime = 'DateTimeFilter'
    Boolean = 'BooleanFilter'
    Integer = 'IntFilter'
    String = 'StringFilter'
    RefFilter = 'RefFilter'
    IntRefFilter = 'IntRefFilter'
    BooleanRefFilter = 'BooleanRefFilter'

    @staticmethod
    def fromValue(value, ref):
        if isinstance(value, datetime):
            return FilterType.DateTime
        if isinstance(value, bool):
            return FilterType.BooleanRefFilter if ref else FilterType.Boolean
        if isinstance(value, int):
            return FilterType.IntRefFilter if ref else FilterType.Integer
        if isinstance(value, str):
            return FilterType.RefFilter if ref else FilterType.String
        raise TypeError("value must be of one of the following types: datetime, bool, int or str")


class Filter(Object):

    def __init__(self, field):
        super().__init__()
        self.field = field


class FilterBuilder(Object):

    def __init__(self, name, reference=False):
        super().__init__()
        self.filter = Filter(name)
        self.reference = reference

    @staticmethod
    def ref(name):
        return FilterBuilder(name, True)

    def like(self, value):
        self.filter.restriction = Restriction.LIKE  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def notLike(self, value):
        self.filter.restriction = Restriction.UNLIKE  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def eq(self, value):
        self.filter.restriction = Restriction.EQUALS  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def ne(self, value):
        self.filter.restriction = Restriction.NOT_EQUALS  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def ge(self, value):
        self.filter.restriction = Restriction.GREATER_EQUALS  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def gt(self, value):
        self.filter.restriction = Restriction.GREATER_THAN  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def le(self, value):
        self.filter.restriction = Restriction.LESS_EQUALS  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def lt(self, value):
        self.filter.restriction = Restriction.LESS_THAN  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def before(self, value):
        self.filter.restriction = Restriction.LESS_THAN  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def after(self, value):
        self.filter.restriction = Restriction.GREATER_THAN  # pylint: disable=attribute-defined-outside-init
        return self.setValue(value)

    def setValue(self, value):
        self.filter._classname = FilterType.fromValue(value, self.reference)  # pylint: disable=protected-access,W0201
        if isinstance(value, datetime):
            value = value.strftime('%Y-%m-%dT%H:%M:%S')
        self.filter.value = value  # pylint: disable=attribute-defined-outside-init

        return self.filter


class QueryParams(Object):

    def __init__(self):
        super().__init__()
        self.startFrom = 0
        self.countLimit = 50

    def include_classname(self):
        self._classname = self.__class__.__name__  # pylint: disable=attribute-defined-outside-init

    def increment(self):
        self.startFrom = self.startFrom + self.countLimit


class QueryParamBuilder:

    def __init__(self):
        self.param = QueryParams()

    def include_classname(self):
        self.param.include_classname()
        return self

    def startFrom(self, startFrom):
        self.param.startFrom = startFrom
        return self

    def countLimit(self, countLimit):
        self.param.countLimit = countLimit
        return self

    def include(self, include):
        self.param.include = include  # pylint: disable=attribute-defined-outside-init
        return self

    def addFilter(self, query_filter):
        if not hasattr(self.param, 'filters'):
            self.param.filters = []  # pylint: disable=attribute-defined-outside-init
        self.param.filters.append(query_filter)
        return self

    def orFilter(self, orFilter):
        self.param.orFilter = orFilter  # pylint: disable=attribute-defined-outside-init
        return self

    def sortBy(self, sortBy):
        self.param.sortBy = sortBy  # pylint: disable=attribute-defined-outside-init
        return self

    def allPortals(self, allPortals):
        self.param.allPortals = allPortals  # pylint: disable=attribute-defined-outside-init
        return self

    def ownedBy(self, ownedBy):
        self.param.ownedBy = ownedBy  # pylint: disable=attribute-defined-outside-init
        return self

    def put(self, key, value):
        setattr(self.param, key, value)
        return self

    def build(self):
        return self.param



import logging
from .enum import Mode
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Antivirus(BaseCommand):
    """Edge Filer Antivirus APIs"""

    def __init__(self, edge):
        super().__init__(edge)
        self.settings = Settings(self._edge)

    def enable(self):
        """
        Enable Bit Defender antivirus.
        """
        logger.info('Enabling antivirus.')
        response = self._edge.api.put('/config/av/realtime/mode', Mode.Enabled)
        logger.info('Antivirus enabled.')
        return response

    def disable(self):
        """
        Disable Bit Defender antivirus.
        """
        logger.info('Disabling antivirus.')
        response = self._edge.api.put('/config/av/realtime/mode', Mode.Disabled)
        logger.info('Antivirus disabled.')
        return response

    def update(self):
        """
        Check for updates.
        """
        return self._edge.api.execute('/config/av/updates', 'updatenow')

    def status(self):
        """
        Get Status.
        """
        return self._edge.api.get('/status/av')


class Settings(BaseCommand):

    def get(self):
        """
        Get antivirus settings.
        """
        return self._edge.api.get('/config/av/updates')

    def update(self, schedule, disabled=False):
        """
        Update antivirus settings.

        :param cterasdk.edge.types.AntivirusUpdateSchedule schedule: Antivirus update schedule
        :param bool,optional disabled: Enable or disable automatic updates, defaults to ``False``
        """
        settings = self.get()
        settings.mode = Mode.Disabled if disabled is True else Mode.Enabled
        settings.schedule = schedule
        return self._edge.api.put('/config/av/updates', settings)



import re
import socket
import logging
import ipaddress

from datetime import datetime
from packaging.version import parse as parse_version

from .object import Object
from .enum import DayOfWeek


logger = logging.getLogger('cterasdk.common')


def union(g1, g2):
    """
    Create a union of two lists, without duplicates.

    :return list: the union of the two lists
    """
    return g1 + list(set(g2) - set(g1))


def merge(d1, d2):
    """
    Merge two dictionaries. In case of duplicate keys, this function will return values from the 2nd dictionary

    :return dict: the merged dictionary
    """
    d3 = {}
    if d1:
        d3.update(d1)
    if d2:
        d3.update(d2)
    return d3


class DataUnit:
    """
    Data Unit

    :ivar str PB: Petabytes
    :ivar str TB: Terabytes
    :ivar str GB: Gigabytes
    :ivar str MB: Megabytes
    :ivar str KB: Kilobytes
    :ivar str B: Bytes
    """
    PB = 'PB'
    TB = 'TB'
    GB = 'GB'
    MB = 'MB'
    KB = 'KB'
    B = 'B'


def convert_size(numeric, u1, u2):
    """
    Convert size between data units

    :param str u1: Convert from data unit
    :param str u2: Convert to data unit
    """
    data_units = [DataUnit.B, DataUnit.KB, DataUnit.MB, DataUnit.GB, DataUnit.TB, DataUnit.PB]
    if u1 not in data_units:
        raise ValueError(f"Invalid current unit type {u1}")
    if u2 not in data_units:
        raise ValueError(f"Invalid target unit type {u2}")
    offset = (data_units.index(u1) - data_units.index(u2)) * 10
    if offset > 0:
        return numeric * (1 << offset)
    if offset < 0:
        return numeric / (1 << abs(offset))
    return numeric


def df_military_time(time):
    """
    Format datetime object to military time

    :param datetime.datetime datetime: Datetime object to convert
    :returns: Military time formatted string
    :rtype: str
    """
    if not isinstance(time, datetime):
        raise ValueError(f"Invalid type '{type(time)}', expected 'datetime'")
    return time.strftime('%H:%M:%S')


def day_of_week(day):
    """
    Interpret the day of week from numeric to str

    :param int day: A numeric integer representing the day of week
    :returns: Name of the day
    :rtype: str
    """
    name = {v: k for k, v in DayOfWeek.__dict__.items() if isinstance(v, int)}.get(day)
    if not name:
        raise ValueError(f'Invalid day of week: {day}')
    return name


class BaseObjectRef(Object):
    """
    Class Representing a Portal Base Object Reference

    :ivar str classname: Base object class name
    :ivar str uid: Base object unique identifier (uid)
    :ivar str tenant: Base object tenant name
    :ivar str name: Base object name
    """
    def __init__(self, uid, tenant=None, classname=None, name=None, more=None):
        """
        :param str uid: Base object unique identifier
        :param str,optional tenant: Base object tenant
        :param str,optional classname: Base object class name
        :param str,optional name: Base object name
        :param str,optional more: Base object more info
        """
        super().__init__()
        self.uid = uid
        self.tenant = tenant
        self.classname = classname
        self.name = name
        self.more = more

    def in_tenant_context(self):
        """
        Returns True if the base object exists in a context of a tenant
        """
        return bool(self.tenant)

    def __str__(self):
        """
        Returns a string representation of the base object reference
        """
        return '/'.join(['objs', self.uid, self.tenant or '', self.classname or '', self.name or '', self.more or ''])


def parse_base_object_ref(base_object_ref):
    if not base_object_ref.startswith('objs/'):
        logger.error('Invalid base object reference. %s', {'ref': base_object_ref})
        return None
    base_object_ref = base_object_ref[5:]
    p = re.compile('[^/]*')
    components = ['uid', 'tenant', 'classname', 'name', 'more']
    arguments = {}
    while base_object_ref:
        match = p.match(base_object_ref)
        arguments[components.pop(0)] = match.group()
        base_object_ref = base_object_ref[match.end() + 1:]
    return BaseObjectRef(**arguments)


def parse_to_ipaddress(address):
    """
    Parse an ip or network address into ipaddress object

    :param str address: ip (10.0.0.5) or network address (192.168.44.0/28)
    :return: ipaddress.IPV4Address/IPV6Address or ipaddress.IPV4Network/IPV6Network
    """
    try:
        try:
            ip_addrr = ipaddress.ip_address(address)
            logger.debug('ip address validated. %s', {'ip': str(ip_addrr)})
            return ip_addrr
        except (ValueError, TypeError):
            ip_network = ipaddress.ip_network(address)
            logger.debug('ip network validated. %s', {'network': str(ip_network)})
            return ip_network
    except (ValueError, TypeError):
        err = ValueError(f'{address} does not appear to be an IPv4 or IPv6 network or ip address')
        logger.error('Incorrect entry, please use IPv4 or IPv6 CIDR Formats. %s', {'Error': err})
        raise err


class Version:
    """Software Version"""

    def __init__(self, version):
        self._version = parse_version(version)

    def __eq__(self, v):
        return self._version == parse_version(v)

    def __gt__(self, v):
        return self._version > parse_version(v)

    def __ge__(self, v):
        return self._version >= parse_version(v)

    def __lt__(self, v):
        return self._version < parse_version(v)

    def __le__(self, v):
        return self._version <= parse_version(v)

    def __ne__(self, v):
        return self._version != parse_version(v)

    @property
    def version(self):
        return str(self._version)

    def __str__(self):
        return self.version


def utf8_decode(message):
    """
    Decode UTF-8 String
    """
    return message.decode('utf-8')


def utf8_encode(message):
    """
    Encode UTF-8 String
    """
    return message.encode('utf-8')


def tcp_connect(host, port):
    logger.debug('Testing connection. %s', {'host': host, 'port': port})
    message = f"Connection error to remote host {host} on port {port}."
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    rc = None
    try:
        rc = sock.connect_ex((host, port))
    except socket.gaierror:
        logger.debug(message)
        raise ConnectionError(message)

    if rc != 0:
        logger.debug(message)
        raise ConnectionError(message)
    logger.debug("Connection established to remote host %s on port %s", host, port)



import cterasdk.settings
from ...clients import clients
from ..services import Management
from ..endpoints import EndpointBuilder
from ...lib.session.edge import Session
from ...edge import backup, cli, logs, services, support, sync


class Clients:

    def __init__(self, drive, Portal):
        if Portal:
            drive._Portal = Portal
            drive.default.close()
            drive._ctera_session.start_remote_session(Portal.session())
            self.api = Portal.default.clone(clients.API, EndpointBuilder.new(drive.base), authenticator=lambda *_: True)
        else:
            self.api = drive.default.clone(clients.API, EndpointBuilder.new(drive.base, '/admingui/api'))


class Drive(Management):

    def __init__(self, host=None, port=None, https=True, Portal=None, *, base=None):
        super().__init__(host, port, https, base, cterasdk.settings.drive.syn.settings)
        self._ctera_session = Session(self.host())
        self._ctera_clients = Clients(self, Portal)
        self.backup = backup.Backup(self)
        self.cli = cli.CLI(self)
        self.logs = logs.Logs(self)
        self.services = services.Services(self)
        self.support = support.Support(self)
        self.sync = sync.Sync(self)

    @property
    def api(self):
        return self.clients.api

    @property
    def _session_id_key(self):
        return '_cteraSessionId_'

    @property
    def _login_object(self):
        raise NotImplementedError("Logins to the 'Drive App' are not enabled.")

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['backup', 'cli', 'logs', 'services', 'support', 'sync']



from .base import CTERAException


class RemoteStorageError(CTERAException):
    """
    Base Exception for Remote File Storage

    :ivar str path: Path
    """
    def __init__(self, message, path=None):
        super().__init__(message)
        self.path = path


class ResourceNotFoundError(RemoteStorageError):

    def __init__(self, path):
        super().__init__('Remote directory not found. Please verify the path and try again.', path)


class NotADirectory(RemoteStorageError):

    def __init__(self, path):
        super().__init__('Target validation error: Resource exists but it is not a directory.', path)


class ResourceExistsError(RemoteStorageError):

    def __init__(self, path):
        super().__init__('Resource already exists: a file or folder with this name already exists.', path)


class PathValidationError(RemoteStorageError):

    def __init__(self, path=None, **kwargs):  # pylint: disable=unused-argument
        super().__init__('Path validation failed: the specified destination path does not exist.', path)


class NameSyntaxError(RemoteStorageError):

    def __init__(self, path):
        super().__init__('Invalid name: the name contains characters that are not allowed "\\ / : ? & < > \" |".', path)


class ReservedNameError(RemoteStorageError):

    def __init__(self, path):
        super().__init__('Reserved name error: the name is reserved and cannot be used.', path)


class RestrictedPathError(RemoteStorageError):

    def __init__(self, path):
        super().__init__('Creating a folder in the specified location is forbidden.', path)


class RestrictedRoot(RemoteStorageError):

    def __init__(self):
        super().__init__('Storing files to the root directory is forbidden.', '/')


class PermissionDenied(RemoteStorageError):

    def __init__(self, action, path=None):
        super().__init__('Permission denied: Inappropriate permissions to access this resource.', path)
        self.action = action


class UnwriteableScope(RemoteStorageError):

    def __init__(self, path, scope):
        super().__init__("Write access denied. This target is protected and cannot be modified.", path)
        self.scope = scope


class FileConflict(RemoteStorageError):

    def __init__(self, action, name, cursor):
        super().__init__('Conflict: a file with the same name already exists.')
        self.action = action
        self.name = name
        self.cursor = cursor


class UploadException(RemoteStorageError):
    """
    Upload Exception

    :ivar str path: Path
    """
    def __init__(self, message, path):
        super().__init__(f'Upload failed: {message}.', path)


class OutOfQuota(UploadException):

    def __init__(self, entity, path):
        super().__init__(f'{entity} is out of quota', path)


class RejectedByPolicy(UploadException):

    def __init__(self, path):
        super().__init__('Rejected by Cloud Drive policy rule', path)


class NoStorageBucket(UploadException):

    def __init__(self, path):
        super().__init__('No available storage location', path)


class WindowsACLError(UploadException):

    def __init__(self, path):
        super().__init__('Unable to store file in a Windows ACL-enabled cloud folder', path)



import logging
from abc import abstractmethod


logger = logging.getLogger('cterasdk.common')


class BaseIterator:
    """Abstract Iterator"""

    def __init__(self, callback, parameter):
        self._callback = callback
        self._parameter = parameter
        self._more = True
        self._objects = []

    def __iter__(self):
        return self

    def __next__(self):
        if self._objects:
            return self.object
        if self._more:
            self._more, page = self.page()
            self._objects.extend(page)
            if self._objects:
                return self.object
        logger.debug('Stopping iteration.')
        raise StopIteration

    @property
    def object(self):
        return self._objects.pop(0)

    @abstractmethod
    def page(self):
        raise NotImplementedError("Subclass must implemenet the 'page' function")


class KeyValueQueryIterator(BaseIterator):
    """ Key Value Iterator """

    def page(self):
        objects = self._callback(self._parameter)
        return False, objects


class QueryIterator(BaseIterator):
    """ Objects Iterator """

    def page(self):
        response = self._callback(self._parameter)
        self._parameter.increment()
        return response.more, response.objects


class BaseResponse:

    def __init__(self, response):
        self._response = response

    @property
    @abstractmethod
    def more(self):
        raise NotImplementedError("Subclass must implement the 'more' method")

    @property
    @abstractmethod
    def objects(self):
        raise NotImplementedError("Subclass must implement the 'objects' method")


class DefaultResponse(BaseResponse):

    @property
    def more(self):
        return self._response.hasMore

    @property
    @abstractmethod
    def objects(self):
        return self._response.objects


class QueryLogsResponse(DefaultResponse):

    @property
    def objects(self):
        return self._response.logs


class CursorResponse(BaseResponse):

    @property
    def more(self):
        return self._response.has_more

    @property
    def objects(self):
        return self._response.entries

    @property
    def cursor(self):
        return self._response.cursor



import asyncio
import cterasdk.settings

from . import filters
from .credentials import KeyPair, Bearer, create_bearer_token
from .lib import get_chunks, decrypt_encryption_key, process_chunks
from .types import ByteRange
from .stream import Streamer

from ..objects.endpoints import DefaultBuilder, EndpointBuilder
from ..clients.clients import AsyncClient, AsyncJSON


class DirectIO:

    async def __aenter__(self):
        return self

    def __init__(self, baseurl=None, access_key_id=None, secret_access_key=None, bearer=None):
        """
        Initialize a CTERA Direct IO Client.

        :param str baseurl: Portal URL
        :param str,optional access_key_id: Access key
        :param str,optional secret_access_key: Secret key
        :param str,optional bearer: Bearer token
        """
        self._api = AsyncJSON(EndpointBuilder.new(baseurl, '/directio'), settings=cterasdk.settings.io.direct.api.settings,
                              authenticator=lambda *_: True)
        self._client = AsyncClient(DefaultBuilder(), settings=cterasdk.settings.io.direct.storage.settings, authenticator=lambda *_: True)
        self._credentials = Bearer(bearer) if bearer else KeyPair(access_key_id, secret_access_key)
        self._bearer = create_bearer_token(self._credentials)

    async def _chunks(self, file_id):
        metadata = await get_chunks(self._api, self._bearer, file_id)
        if metadata.encrypted:
            metadata.encryption_key = decrypt_encryption_key(
                metadata.file_id,
                metadata.encryption_key,
                self._credentials.secret_access_key
            )
        return metadata

    async def metadata(self, file_id):
        """
        Get File Metadata.

        :param int file_id: File ID.
        """
        meta = await self._chunks(file_id)
        return meta.serialize()

    async def blocks(self, file_id, byte_range=None, max_workers=None):
        """
        Blocks API.

        :param int file_id: File ID.
        :param cterasdk.direct.types.ByteRange,optional byte_range: Byte Range.
        :param int,optional max_workers: Max concurrent tasks
        :returns: List of Blocks.
        :rtype: list[cterasdk.direct.types.Block]
        """
        meta = await self._chunks(file_id)
        byte_range = byte_range if byte_range is not None else ByteRange.default()
        executor = self.executor(filters.span(meta, byte_range), meta.encryption_key, meta.file_id, max_workers)
        return await executor()

    async def streamer(self, file_id, byte_range=None, max_workers=None):
        """
        Stream API.

        :param int file_id: File ID.
        :param cterasdk.direct.types.ByteRange,optional byte_range: Byte Range.
        :param int,optional max_workers: Max concurrent tasks
        :returns: Streamer Object
        :rtype: cterasdk.direct.stream.Streamer
        """
        meta = await self._chunks(file_id)
        byte_range = byte_range if byte_range is not None else ByteRange.default()
        max_workers = max_workers if max_workers else cterasdk.settings.io.direct.streamer.max_workers
        executor = self.executor(filters.span(meta, byte_range), meta.encryption_key, file_id, max_workers)
        return Streamer(executor, byte_range)

    def executor(self, chunks, encryption_key, file_id=None, max_workers=None):
        """
        Download Executor.

        :param list[cterasdk.direct.types.Chunk] chunks: List of Chunks.
        :param str encryption_key: Decryption Key.
        :param int,optional file_id: File ID.
        :param int,optional max_workers: Max concurrent tasks.

        :returns: Callable Downloader
        :rtype: function
        """

        async def execute():
            """
            Asynchronous Executable of Chunk Retrieval Tasks.
            """
            return await process_chunks(self._client, file_id, chunks, encryption_key,
                                        asyncio.Semaphore(max_workers) if max_workers else None)

        return execute

    async def close(self):
        await self._api.close()
        await self._client.close()

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()



from .base import CTERAException


class AwaitableTaskException(CTERAException):

    def __init__(self, message, awaitable_task):
        """
        Awaitable Task Exception

        :param cterasdk.lib.tasks.AwaitableTask awaitable_task: Awaitable task object
        """
        super().__init__(message)
        self.awaitable_task = awaitable_task


class TaskException(CTERAException):
    """
    Task Exception

    :param cterasdk.common.object.Object task: Task object
    """
    def __init__(self, message, task):
        super().__init__(message)
        self.task = task


class TaskWaitTimeoutError(TaskException):
    """
    Task Wait Timeout Error
    """
    def __init__(self, duration, task):
        super().__init__(f"Task {task.id} remains pending completion after {duration} second(s).", task)



from abc import abstractmethod

from . import endpoints, uri
from .utils import URI
from ..clients import clients
from ..common import Object
from ..convert import tojsonstr


class Service:

    def __init__(self, host, port, https, base):
        self._base = base if base else str(URI.instance(f'http{"s" if https else ""}', host, port or 443 if https else 80))

    @property
    def base(self):
        return self._base

    def host(self):
        hostname = uri.components(self._base).hostname
        return hostname if hostname else ''

    def port(self):
        components = uri.components(self._base)
        return components.port or 443 if components.scheme == 'https' else 80

    @property
    def _omit_fields(self):
        return []

    def __str__(self):
        x = Object()
        x.__dict__ = {k: v for k, v in self.__dict__.items() if not (k.startswith('_') or k in self._omit_fields)}
        return tojsonstr(x)


class CTERA(Service):

    def __init__(self, host, port, https, base):
        super().__init__(host, port, https, base)
        self._ctera_session = None
        self._ctera_clients = None

    @property
    def clients(self):
        return self._ctera_clients

    def session(self):
        return self._ctera_session

    def _before_login(self):
        """Override to implement logic before login"""

    def _after_login(self):
        """Override to implement logic after login"""

    @property
    @abstractmethod
    def _login_object(self):
        raise NotImplementedError(
            "Implementing class must implement the login_object property by returning an object with login and logout methods"
        )

    @abstractmethod
    def _authenticator(self, url):
        raise NotImplementedError("Subclass must implement the '_authenticator' function")

    def whoami(self):
        """
        Return the name of the logged in user.

        :return cterasdk.common.object.Object: The session object of the current user
        """
        return self._ctera_session.whoami()


class AsyncManagement(CTERA):  # pylint: disable=abstract-method

    async def __aenter__(self):
        return self

    def __init__(self, host, port, https, base, settings):
        super().__init__(host, port, https, base)
        self._default = clients.AsyncClient(endpoints.EndpointBuilder.new(self.base), settings=settings, authenticator=self._authenticator)

    @property
    def default(self):
        return self._default

    async def login(self, username, password):
        self._before_login()
        await self._login_object.login(username, password)
        await self._ctera_session.async_start_session(self)
        self._after_login()

    async def logout(self):
        if self._ctera_session.connected:
            await self._login_object.logout()
            self._ctera_session.stop_session()
        await self.default.close()

    async def __aexit__(self, exc_type, exc, tb):
        await self.default.close()


class Management(CTERA):

    def __enter__(self):
        return self

    def __init__(self, host, port, https, base, settings):
        super().__init__(host, port, https, base)
        self._default = clients.Client(endpoints.EndpointBuilder.new(self.base), settings=settings, authenticator=self._authenticator)

    def login(self, username, password):
        """
        Log in

        :param str username: Username
        :param str password: Password
        """
        self._before_login()
        self._login_object.login(username, password)
        self._ctera_session.start_session(self)
        self._after_login()

    def logout(self):
        """ Log out """
        if self._ctera_session.connected:
            self._login_object.logout()
            self._ctera_session.stop_session()
        self._default.close()

    @abstractmethod
    def test(self):
        return NotImplementedError("Subclass must implement the 'test' function")

    @property
    @abstractmethod
    def _session_id_key(self):
        return NotImplementedError("Subclass must implement the '_session_id_key' property")

    @property
    def default(self):
        return self._default

    def get_session_id(self):
        """
        Get Session Identifier

        :return str: Session ID
        """
        return self._default.cookies.get(self._session_id_key)

    def set_session_id(self, session_id):
        self._default.cookies.update({self._session_id_key: session_id}, self._default.baseurl)
        self._ctera_session.start_session(self)

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._default.close()



import logging
import aiofiles
from .commonfs import write_new_version, ResultContext


logger = logging.getLogger('cterasdk.filesystem')


async def write(directory, name, handle):
    """
    Write bytes to disk.

    :param str directory: Directory
    :param str name: Name
    :param bytes handle: Handle
    :returns: Path
    :rtype: str
    """
    ctx = ResultContext()
    with write_new_version(directory, name) as tempfile:
        await overwrite(tempfile, handle)
    return ctx.value


async def overwrite(p, handle):
    """
    Write, without validation.

    :param Path p: Path
    :param bytes handle: Handle
    """
    async with aiofiles.open(p, 'w+b') as fd:
        if isinstance(fd, bytes):
            await fd.write(handle)
        else:
            async for chunk in handle.a_iter_content(chunk_size=8192):
                await fd.write(chunk)
    return p.as_posix()



import logging
import platform
import threading
import aiohttp
from . import async_requests, errors
from .settings import ClientSessionSettings, TraceSettings
from ..common import utils
from ..version import __version__


logger = logging.getLogger('cterasdk.http')


class CookieJar:

    def __init__(self, cookies):
        self._cookies = cookies

    def get(self, key):
        return self._cookies.get(key)

    def update(self, cookies, response_url):
        self._cookies.update_cookies(cookies, response_url)

    def clear(self):
        self._cookies.clear()


class PersistentHeaders:
    """Headers to include in every request"""

    def __init__(self):
        self._headers = {
            'User-Agent': (
                f"CTERA Python SDK/{__version__}; aiohttp/{aiohttp.__version__};"
                f" ({' '.join([platform.system(), str(platform.release())])}; {platform.machine()};"
                f" Python {'.'.join(platform.python_version_tuple())});"
            )
        }

    @property
    def all(self):
        return self._headers

    def persist_response_header(self, response, header):
        """
        Persist header from response object.

        :param cterasdk.clients.base.BaseResponse response: Response object
        :param str header: Header name
        """
        value = response.headers.get(header, None)
        if value:
            self.persist_headers({header: value})
        else:
            logger.debug("Could not find header: '%s' in response", header)

    def persist_headers(self, headers):
        self._headers.update(headers)


class BaseClient:
    """Base Client"""

    def __init__(self, builder=None, session=None, settings=None, authenticator=None):
        """
        Initialize a Client

        :param builder: Endpoint builder.
        :param ,optional session: Session.
        :param ,optional settings: Client Session Settings.
        :param ,optional authenticator: Authenticator function.
        """
        self._headers = PersistentHeaders()
        self._authenticator = authenticator
        self._builder = builder

        default_settings = ClientSessionSettings()
        if settings:
            default_settings.update(**settings)

        self._session = session if session else async_requests.Session(default_settings, TraceSettings())

    def clone(self, definition, builder=None, authenticator=None):
        """
        Clone a Client

        :param class definition: Class definition.
        :param ,optional builder: Endpoint builder.
        :param ,optional authenticator: Authenticator function.
        """
        return definition(
            builder if builder is not None else self._builder,
            self._session,
            None,
            authenticator if authenticator is not None else self._authenticator,
        )

    @property
    def cookies(self):
        return CookieJar(self._session.cookies)

    @property
    def headers(self):
        return self._headers

    def join_headers(self, request):
        request.kwargs['headers'] = utils.merge(request.kwargs.get('headers', None), self.headers.all)
        return request

    @property
    def baseurl(self):
        return self._builder()

    def request(self, request, *, on_response=None, on_error=None):
        on_error = on_error if on_error else errors.DefaultHandler()
        return self._request(request, on_response=on_response, on_error=on_error)

    async def a_request(self, request, *, on_response=None, on_error=None):
        on_error = on_error if on_error else errors.DefaultHandler()
        return await self._request(request, on_response=on_response, on_error=on_error)

    async def close(self):
        await self._session.close()

    def __str__(self):
        return f"({self.__class__.__name__} client at {hex(hash(self))}, baseurl={self.baseurl})"


class BaseResponse:
    """Base Response Object"""

    def __init__(self, response):
        self._response = response

    @property
    def method(self):
        return self._response.method

    @property
    def ok(self):
        return self._response.ok

    @property
    def reason(self):
        return self._response.reason

    @property
    def url(self):
        return self._response.url

    @property
    def real_url(self):
        return self._response.real_url

    @property
    def status(self):
        return self._response.status

    @property
    def headers(self):
        return self._response.headers

    def raise_for_status(self):
        return self._response.raise_for_status()


def run_threadsafe(loop, target, *args, **kwargs):
    event = threading.Event()

    t = Task(loop, event, target, *args, **kwargs)
    t.start()

    event.wait()

    if t.exception:
        raise t.exception
    return t.response


class Task(threading.Thread):

    def __init__(self, loop, event, target, *args, **kwargs):
        super().__init__(name='Thread-safe Executor')
        self.loop = loop
        self.event = event
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.exception = None
        self.response = None

    def run(self):
        try:
            self.response = self.loop.run_until_complete(self.target(*self.args, **self.kwargs))
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.exception = e
        finally:
            self.event.set()



import logging

from ...cio.core import CorePath, await_or_future
from ...exceptions.transport import HTTPError
from ...exceptions.io import FileConflict
from ...lib.storage import synfs, commonfs
from ..base_command import BaseCommand
from . import io


logger = logging.getLogger('cterasdk.core')


class FileBrowser(BaseCommand):

    def __init__(self, core):
        super().__init__(core)
        self._scope = f'/{self._core.context}/webdav'

    def handle(self, path):
        """
        Get File Handle.

        :param str path: Path to a file
        """
        handle_function = io.handle(self.normalize(path))
        return handle_function(self._core)

    def handle_many(self, directory, *objects):
        """
        Get a Zip Archive File Handle.

        :param str directory: Path to a folder
        :param args objects: List of files and folders
        """
        handle_many_function = io.handle_many(self.normalize(directory), *objects)
        return handle_many_function(self._core)

    def download(self, path, destination=None):
        """
        Download a file

        :param str path: Path
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        directory, name = commonfs.determine_directory_and_filename(path, destination=destination)
        handle = self.handle(path)
        return synfs.write(directory, name, handle)

    def download_many(self, target, objects, destination=None):
        """
        Download selected files and/or directories as a ZIP archive.

        .. warning::
            The provided list of objects is not validated. Only existing files and directories
            will be included in the resulting ZIP file.

        :param str target:
            Path to the cloud folder containing the files and directories to download.
        :param list[str] objects:
            List of file and/or directory names to include in the download.
        :param str destination:
            Optional. Path to the destination file or directory. If a directory is provided,
            the original filename will be preserved. Defaults to the default download directory.
        """
        directory, name = commonfs.determine_directory_and_filename(target, objects, destination=destination, archive=True)
        handle = self.handle_many(target, *objects)
        return synfs.write(directory, name, handle)

    def listdir(self, path=None, depth=None, include_deleted=False):
        """
        List Directory

        :param str,optional path: Path, defaults to the Cloud Drive root
        :param bool,optional include_deleted: Include deleted files, defaults to False
        """
        return io.listdir(self._core, self.normalize(path), depth=depth, include_deleted=include_deleted)

    def exists(self, path):
        """
        Check if item exists

        :param str path: Path
        """
        return io.exists(self._core, self.normalize(path))

    def versions(self, path):
        """
        List snapshots of a file or directory

        :param str path: Path
        """
        return io.versions(self._core, self.normalize(path))

    def walk(self, path=None, include_deleted=False):
        """
        Walk Directory Contents

        :param str,optional path: Path to walk, defaults to the root directory
        :param bool,optional include_deleted: Include deleted files, defaults to False
        """
        return io.walk(self._core, self._scope, path, include_deleted=include_deleted)

    def public_link(self, path, access='RO', expire_in=30):
        """
        Create a public link to a file or a folder

        :param str path: The path of the file to create a link to
        :param str,optional access: Access policy of the link, defaults to 'RO'
        :param int,optional expire_in: Number of days until the link expires, defaults to 30
        """
        return io.public_link(self._core, self.normalize(path), access, expire_in)

    def _try_with_resolver(self, func, *paths, destination=None, resolver=None, cursor=None, wait=True):
        def wrapper(resume_from=None):
            ref = func(self._core, *paths, destination=destination, resolver=resolver, cursor=resume_from)
            return await_or_future(self._core, ref, wait)

        try:
            return wrapper(cursor)
        except FileConflict as e:
            if resolver:
                return wrapper(e.cursor)
            raise

    def copy(self, *paths, destination=None, resolver=None, cursor=None, wait=True):
        """
        Copy one or more files or folders

        :param list[str] paths: List of paths
        :param str destination: Destination
        :param cterasdk.core.types.ConflictResolver resolver: Conflict resolver, defaults to ``None``
        :param cterasdk.common.object.Object cursor: Resume copy from cursor
        :param bool,optional wait: ``True`` Wait for task to complete, or ``False`` to return an awaitable task object.
        :returns: Task status object, or an awaitable task object
        :rtype: cterasdk.common.object.Object or :class:`cterasdk.lib.tasks.AwaitablePortalTask`
        """
        try:
            return self._try_with_resolver(io.copy, *[self.normalize(path) for path in paths],
                                           destination=self.normalize(destination),
                                           resolver=resolver, cursor=cursor, wait=wait)
        except ValueError:
            raise ValueError('Copy destination was not specified.')

    def permalink(self, path):
        """
        Get Permalink for Path.

        :param str path: Path.
        """
        p = self.normalize(path)
        for e in io.listdir(self._core, p.parent, 1, False, p.name, 1):
            if e.name == p.name:
                return e.permalink
        raise FileNotFoundError('File not found.', path)

    def normalize(self, entries):
        return CorePath.instance(self._scope, entries)


class CloudDrive(FileBrowser):

    def upload(self, name, destination, handle, size=None):
        """
        Upload from file handle.

        :param str name: File name.
        :param str destination: Path to remote directory.
        :param object handle: File handle, String, or Bytes.
        :param str,optional size: File size, defaults to content length
        """
        upload_function = io.upload(name, size, self.normalize(destination), handle)
        return upload_function(self._core)

    def upload_file(self, path, destination):
        """
        Upload a file.

        :param str path: Local path
        :param str destination: Remote path
        """
        with open(path, 'rb') as handle:
            metadata = commonfs.properties(path)
            response = self.upload(metadata['name'], destination, handle, metadata['size'])
        return response

    def mkdir(self, path):
        """
        Create a new directory

        :param str path: Directory path
        """
        return io.mkdir(self._core, self.normalize(path))

    def makedirs(self, path):
        """
        Create a directory recursively

        :param str path: Directory path
        """
        return io.makedirs(self._core, self.normalize(path))

    def rename(self, path, name, *, wait=True):
        """
        Rename a file

        :param str path: Path of the file or directory to rename
        :param str name: The name to rename to
        :param bool,optional wait: ``True`` Wait for task to complete, or ``False`` to return an awaitable task object.
        :returns: Task status object, or an awaitable task object
        :rtype: cterasdk.common.object.Object or :class:`cterasdk.lib.tasks.AwaitablePortalTask`
        """
        ref = io.rename(self._core, self.normalize(path), name)
        return await_or_future(self._core, ref, wait)

    def delete(self, *paths, wait=True):
        """
        Delete one or more files or folders

        :param str path: Path
        :param bool,optional wait: ``True`` Wait for task to complete, or ``False`` to return an awaitable task object.
        :returns: Task status object, or an awaitable task object
        :rtype: cterasdk.common.object.Object or :class:`cterasdk.lib.tasks.AwaitablePortalTask`
        """
        ref = io.remove(self._core, *[self.normalize(path) for path in paths])
        return await_or_future(self._core, ref, wait)

    def undelete(self, *paths, wait=True):
        """
        Recover one or more files or folders

        :param str path: Path
        :param bool,optional wait: ``True`` Wait for task to complete, or ``False`` to return an awaitable task object.
        :returns: Task status object, or an awaitable task object
        :rtype: cterasdk.common.object.Object or :class:`cterasdk.lib.tasks.AwaitablePortalTask`
        """
        ref = io.recover(self._core, *[self.normalize(path) for path in paths])
        return await_or_future(self._core, ref, wait)

    def move(self, *paths, destination=None, resolver=None, cursor=None, wait=True):
        """
        Move one or more files or folders

        :param list[str] paths: List of paths
        :param str destination: Destination
        :param cterasdk.core.types.ConflictResolver resolver: Conflict resolver, defaults to ``None``
        :param cterasdk.common.object.Object cursor: Resume copy from cursor
        :param bool,optional wait: ``True`` Wait for task to complete, or ``False`` to return an awaitable task object.
        :returns: Task status object, or an awaitable task object
        :rtype: cterasdk.common.object.Object or :class:`cterasdk.lib.tasks.AwaitablePortalTask`
        """
        try:
            return self._try_with_resolver(io.move, *[self.normalize(path) for path in paths],
                                           destination=self.normalize(destination),
                                           resolver=resolver, cursor=cursor, wait=wait)
        except ValueError:
            raise ValueError('Move destination was not specified.')

    def get_share_info(self, path):
        """
        Get share settings and recipients

        :param str path: Path
        """
        return io.get_share_info(self._core, self.normalize(path))

    def share(self, path, recipients, as_project=True, allow_reshare=True, allow_sync=True):
        """
        Share a file or a folder

        :param str path: The path of the file or folder to share
        :param list[cterasdk.core.types.ShareRecipient] recipients: A list of share recipients
        :param bool,optional as_project: Share as a team project, defaults to True when the item is a cloud folder else False
        :param bool,optional allow_reshare: Allow recipients to re-share this item, defaults to True
        :param bool,optional allow_sync: Allow recipients to sync this item, defaults to True when the item is a cloud folder else False
        :return: A list of all recipients added to the collaboration share
        :rtype: list[cterasdk.core.types.ShareRecipient]
        """
        return io.share(self._core, self.normalize(path), recipients, as_project, allow_reshare, allow_sync)

    def add_share_recipients(self, path, recipients):
        """
        Add share recipients

        :param str path: The path of the file or folder
        :param list[cterasdk.core.types.ShareRecipient] recipients: A list of share recipients
        :return: A list of all recipients added
        :rtype: list[cterasdk.core.types.ShareRecipient]
        """
        return io.add_share_recipients(self._core, self.normalize(path), recipients)

    def remove_share_recipients(self, path, accounts):
        """
        Remove share recipients

        :param str path: The path of the file or folder
        :param list[cterasdk.core.types.PortalAccount] accounts: A list of portal user or group accounts
        :return: A list of all share recipients removed
        :rtype: list[cterasdk.core.types.PortalAccount]
        """
        return io.remove_share_recipients(self._core, self.normalize(path), accounts)

    def unshare(self, path):
        """
        Unshare a file or a folder
        """
        return io.unshare(self._core, self.normalize(path))


class Backups(FileBrowser):

    def device_config(self, device, destination=None):
        """
        Download a device configuration file

        :param str device: The device name
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        try:
            destination = destination if destination is not None else f'{commonfs.downloads()}/{device}.xml'
            return self.download(f'backups/{device}/Device Configuration/db.xml', destination)
        except HTTPError as error:
            logger.error('Error downloading device configuration: %s. Reason: %s', device, error.response.reason)
            raise error



import re
import copy
import logging

from .base_command import BaseCommand
from . import query, devices
from .enum import ListFilter, PolicyType
from .types import ComplianceSettingsBuilder, ExtendedAttributesBuilder
from ..common import union, Object
from ..exceptions import CTERAException, ObjectNotFoundException


logger = logging.getLogger('cterasdk.core')


class CloudFS(BaseCommand):
    """
    CloudFS APIs

    :ivar cterasdk.core.cloudfs.FolderGroups groups: Object holding Folder Groups APIs
    :ivar cterasdk.core.cloudfs.CloudDrives drives: Object holding Cloud Drive Folders APIs
    :ivar cterasdk.core.cloudfs.Backups backups: Object holding Backup Folders APIs
    :ivar cterasdk.core.cloudfs.Zones zones: Object holding Zones APIs
    :ivar cterasdk.core.cloudfs.Exports exports: Object holding Bucket Exports APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.groups = FolderGroups(self._core)
        self.drives = CloudDrives(self._core)
        self.backups = Backups(self._core)
        self.zones = Zones(self._core)
        self.exports = Exports(self._core)


class FolderGroups(BaseCommand):
    """ Folder Groups APIs """

    default = ['name', 'owner']

    def _get_entire_object(self, name):
        ref = f'/foldersGroups/{name}'
        try:
            return self._core.api.get(ref)
        except CTERAException as error:
            raise CTERAException(f'Folder group not found: {ref}') from error

    def get(self, name, include=None):
        """
        Get folder group

        :param str name: Name of the Folder Group to find
        :param str,optional include: List of fields to retrieve, defaults to ['name', 'owner']
        """
        include = union(include or [], FolderGroups.default)
        include = ['/' + attr for attr in include]
        folder_group = self._core.api.get_multi(f'/foldersGroups/{name}', include)
        if folder_group.name is None:
            raise ObjectNotFoundException(f'/foldersGroups/{name}')
        return folder_group

    def all(self, include=None, user=None):
        """
        List folder groups

        :param str,optional include: List of fields to retrieve, defaults to ['name', 'owner']
        :param cterasdk.core.types.UserAccount user: User account of the folder group owner
        :returns: Iterator for all folder groups
        """
        include = union(include or [], FolderGroups.default)
        builder = query.QueryParamBuilder().include(include)
        if user:
            uid = self._core.users.get(user, ['uid']).uid
            builder.ownedBy(uid)
        param = builder.build()
        return query.iterator(self._core, '/foldersGroups', param)

    def add(self, name, user=None, deduplication_method_type=None, storage_class=None):
        """
        Create a Folder Group

        :param str name: Name of the new folder group
        :param cterasdk.core.types.UserAccount user:
         User account, the user directory and name of the new folder group owner (default to None)
        :param cterasdk.core.enum.DeduplicationMethodType deduplication_method_type: Deduplication-Method
        :param str,optional storage_class: Storage class, defaults to the Default storage class
        """

        param = Object()
        param.name = name
        param.disabled = False
        param.owner = self._core.users.get(user, ['baseObjectRef']).baseObjectRef if user is not None else None
        param.deduplicationMethodType = deduplication_method_type
        if storage_class:
            param.storageClass = self._core.storage_classes.get(storage_class).baseObjectRef

        try:
            response = self._core.api.execute('', 'createFolderGroup', param)
            logger.info('Folder group created. %s', {'name': name, 'owner': param.owner})
            return response
        except CTERAException as error:
            logger.error('Folder group creation failed. %s', {'name': name, 'owner': str(user)})
            raise error

    def modify(self, current_name, new_name):
        """
        Modify a folder group

        :param str current_name: Current folder group name
        :param str new_name: New folder group name
        """
        param = self._get_entire_object(current_name)
        if new_name:
            param.name = new_name

        ref = f'/foldersGroups/{current_name}'
        try:
            response = self._core.api.put(ref, param)
            logger.info('Folder group modified: %s', current_name)
            return response
        except CTERAException as error:
            logger.error('Folder group modification failed: %s', current_name)
            raise CTERAException(f'Folder group modification failed: {ref}') from error

    def delete(self, name):
        """
        Remove a Folder Group

        :param str name: Name of the folder group to remove
        """

        logger.info('Deleting folder group. %s', {'name': name})
        self._core.api.execute('/foldersGroups/' + name, 'deleteGroup', True)
        logger.info('Folder group deleted. %s', {'name': name})


class CloudDrives(BaseCommand):
    """ Cloud Drive Folder APIs

    :ivar cterasdk.core.cloudfs.Locks locks: Global file locking APIs
    """

    default = ['name', 'group', 'owner']

    default_extensions = ["ppt", "pptx", "xls", "xlsx", "doc", "docx", "indd", "idlk", "dwl", "dwl2", "dwt", "dwg", "rvt", "dat"]

    def __init__(self, core):
        super().__init__(core)
        self.locks = Locks(self._core)

    def _get_entire_object(self, name, owner):
        return self._core.api.get(f'{self.find(name, owner, include=["baseObjectRef"]).baseObjectRef}')

    def add(self, name, group, owner, winacls=True, description=None,  # pylint: disable=too-many-arguments
            quota=None, compliance_settings=None, xattrs=None, gfl=False, lock_extensions=None):
        """
        Create a new Cloud Drive Folder (Cloud Volume)

        :param str name: Name of the new folder
        :param str group: Folder Group to assign this folder to
        :param cterasdk.core.types.UserAccount owner: User account, the owner of the new folder
        :param bool,optional winacls: Use Windows ACLs, defaults to True
        :param str,optional description: Cloud drive folder description
        :param str,optional quota: Cloud drive folder quota in GB
        :param cterasdk.common.object.Object,optional compliance_settings: Compliance settings, defaults to disabled.
         Use :func:`cterasdk.core.types.ComplianceSettingsBuilder` to build the compliance settings object
        :param cterasdk.common.object.Object,optional xattrs: Extended attributes, defaults to MacOS.
         Use :func:`cterasdk.core.types.ExtendedAttributesBuilder` to build the extended attributes object
        :param bool,optional gfl: Enable global file locking
        :param list[str],optional lock_extensions: List of file extensions (without leading dot) for which global file locking is enforced.
        :returns: Path to the Cloud Drive folder
        :rtype: str
        """
        param = Object()
        param.name = name
        param.owner = self._core.users.get(owner, ['baseObjectRef']).baseObjectRef
        param.group = self._core.cloudfs.groups.get(group, ['baseObjectRef']).baseObjectRef
        param.enableSyncWinNtExtendedAttributes = winacls
        param.folderQuota = quota
        if description:
            param.description = description
        param.wormSettings = compliance_settings if compliance_settings else ComplianceSettingsBuilder.default().build()
        if xattrs:
            param.extendedAttributes = xattrs
        elif not winacls:  # Only override default when winacls is False
            param.extendedAttributes = Object()
            param.extendedAttributes._classname = 'ExtendedAttributes'  # pylint: disable=protected-access
            param.extendedAttributes.enable = False
            param.extendedAttributes.attributes = [Object()]
            param.extendedAttributes.attributes[0]._classname = 'ExtendedAttributesInfo'  # pylint: disable=protected-access
            param.extendedAttributes.attributes[0].name = 'MacOS'
            param.extendedAttributes.attributes[0].supported = True
        else:
            param.extendedAttributes = ExtendedAttributesBuilder.default().build()

        if gfl:
            param.globalFileLockSettings = Object()
            param.globalFileLockSettings._classname = 'GlobalFileLockSettings'  # pylint: disable=protected-access
            param.globalFileLockSettings.enabled = True
            param.globalFileLockSettings.globalFileLockExtensions = (
                lock_extensions if lock_extensions else CloudDrives.default_extensions
            )

        try:
            response = self._core.api.execute('', 'addCloudDrive', param)
            logger.info(
                'Cloud drive folder created. %s',
                {'name': name, 'owner': param.owner, 'folder_group': group, 'winacls': winacls}
            )
            return re.search(r'/Users\/(.+)', response).group()
        except CTERAException as error:
            logger.error(
                'Cloud drive folder creation failed. %s',
                {'name': name, 'folder_group': group, 'owner': str(owner), 'win_acls': winacls}
            )
            raise error

    def modify(self, current_name, owner, new_name=None,  # pylint: disable=too-many-arguments, too-many-locals
               new_owner=None, new_group=None, description=None, winacls=None, quota=None, compliance_settings=None, xattrs=None,
               gfl=None, lock_extensions=None):
        """
        Modify a Cloud Drive Folder (Cloud Volume)

        :param str current_name: Current folder name
        :param cterasdk.core.types.UserAccount owner: User account, the owner of the folder
        :param str,optional new_name: New folder name
        :param cterasdk.core.types.UserAccount,optional new_owner: User account, the new owner of the folder
        :param str,optional new_group: Folder Group to assign this folder to
        :param str,optional description: Folder description
        :param bool,optional winacls: Enable or disable Windows ACLs
        :param str,optional quota: Folder quota in GB
        :param cterasdk.common.object.Object,optional compliance_settings: Compliance settings.
         Use :func:`cterasdk.core.types.ComplianceSettingsBuilder` to build the compliance settings object
        :param cterasdk.common.object.Object,optional xattrs: Extended attributes.
         Use :func:`cterasdk.core.types.ExtendedAttributesBuilder` to build the extended attributes object
        :param bool,optional gfl: Enable global file locking
        :param list[str],optional lock_extensions: List of file extensions (without leading dot) for which global file locking is enforced.
        """
        param = self._get_entire_object(current_name, owner)
        if new_name:
            param.name = new_name
        if new_owner:
            param.owner = self._core.users.get(new_owner, ['baseObjectRef']).baseObjectRef
        if new_group:
            param.group = self._core.cloudfs.groups.get(new_group, include=['baseObjectRef']).baseObjectRef
        if description:
            param.description = description
        if winacls:
            param.enableSyncWinNtExtendedAttributes = winacls
        if quota:
            param.folderQuota = quota
        if compliance_settings:
            param.wormSettings = compliance_settings
        if xattrs:
            param.extendedAttributes = xattrs
        if gfl is not None:
            param.globalFileLockSettings.enabled = gfl
            if lock_extensions:
                param.globalFileLockSettings.globalFileLockExtensions = (
                    lock_extensions if lock_extensions else CloudDrives.default_extensions
                )

        try:
            response = self._core.api.put(f'/{param.baseObjectRef}', param)
            logger.info('Cloud drive folder updated. %s', {'name': current_name})
            return response
        except CTERAException as error:
            logger.error('Cloud drive folder update failed. %s', {'name': current_name})
            raise error

    def all(self, include=None, list_filter=ListFilter.NonDeleted, user=None):
        """
        List Cloud Drive folders.

        :param str,optional include: List of fields to retrieve, defaults to ['name', 'group', 'owner']
        :param cterasdk.core.enum.ListFilter list_filter: Filter the list of Cloud Drive folders, defaults to non-deleted folders
        :param cterasdk.core.types.UserAccount user: User account of the cloud folder owner
        :returns: Iterator for all Cloud Drive folders
        """
        include = union(include or [], CloudDrives.default)
        builder = query.QueryParamBuilder().include(include)
        if list_filter != ListFilter.NonDeleted:
            builder.put('includeDeleted', True)
            if list_filter == ListFilter.Deleted:
                query_filter = query.FilterBuilder('isDeleted').eq(True)
                builder.addFilter(query_filter)
        if user:
            uid = self._core.users.get(user, ['uid']).uid
            builder.ownedBy(uid)
        param = builder.build()
        return query.iterator(self._core, '/cloudDrives', param)

    def find(self, name, owner, include=None):
        """
        Find a Cloud Drive Folder

        :param str name: Name of the Cloud Drive Folder to find
        :param cterasdk.core.types.UserAccount owner: Cloud drive folder owner
        :param list[str] include: List of metadata fields to include in the response

        :returns: A Cloud Drive Folder
        """
        include = union(include or [], CloudDrives.default)
        builder = query.QueryParamBuilder().include(include)
        builder.addFilter(query.FilterBuilder('name').eq(name))
        builder.addFilter(query.FilterBuilder('owner', True).eq(owner.name))
        builder.put('includeDeleted', True)
        param = builder.build()

        iterator = query.iterator(self._core, '/cloudDrives', param)
        try:
            return next(iterator)
        except StopIteration:
            logger.info('Could not find cloud folder. %s', {'folder': name, 'owner': str(owner)})
            raise CTERAException(f"Unable to locate the Cloud Drive folder '{name}' owned by '{str(owner)}.")

    def delete(self, name, owner, *, permanently=False):
        """
        Delete a Cloud Drive Folder

        :param str name: Name of the Cloud Drive Folder to delete
        :param cterasdk.core.types.UserAccount owner: User account, the owner of the Cloud Drive Folder to delete
        :param bool,optional permanently: Delete permanently
        """
        cloudfolder = self.find(name, owner, include=['uid', 'isDeleted'])
        logger.info('Deleting cloud drive folder. %s', {'name': name, 'owner': str(owner), 'permanently': permanently})
        if permanently:
            return self._core.api.execute(f'/objs/{cloudfolder.uid}', 'deleteFolderPermanently')
        if not cloudfolder.isDeleted:
            return self._core.api.execute(f'/objs/{cloudfolder.uid}', 'delete')
        logger.info('Cloud Drive folder was already deleted. %s', {'name': cloudfolder.name})
        return None

    def recover(self, name, owner):
        """
        Recover a deleted a Cloud Drive Folder

        :param str name: Name of the Cloud Drive Folder to un-delete
        :param cterasdk.core.types.UserAccount owner: User account, the owner of the Cloud Drive Folder to delete
        """
        display_name = self._core.users.get(owner, ['displayName']).displayName
        logger.info('Recovering cloud drive folder. %s', {'name': name, 'owner': str(owner)})
        return self._core.files.undelete(f'Users/{display_name}/{name}')

    def setfacl(self, paths, acl, recursive=False):
        """
        Changing the file or folder ACLs

        :param list(str) paths:  List of folder paths
        :param str acl: Access control list (ACL) represented as an SDDL String
        :param bool,optional recursive: Apply changes recursively to subfolders and files
        """
        param = Object()
        param._classname = 'SDDLFoldersParam'  # pylint: disable=protected-access
        param.foldersPath = paths
        param.sddlString = acl
        param.isRecursive = recursive
        try:
            return self._core.api.execute('', 'setFoldersACL', param)
        except CTERAException as error:
            logger.error('An error occurred attempting to update access control entries.')
            raise CTERAException('ACL modification failed.') from error

    def setoacl(self, paths, owner_sid, recursive=False):
        """
        Changing the File or Folder Owner SID or ACLs

        :param list(str) paths:  List of folder paths
        :param str owner_sid: Owner SID (Security Descriptor)
        :param bool,optional recursive: Apply changes recursively to subfolders and files
        """
        param = Object()
        param._classname = 'OwnerSidFoldersParam'  # pylint: disable=protected-access
        param.foldersPath = paths
        param.ownerSid = owner_sid
        param.isRecursive = recursive
        try:
            return self._core.api.execute('', 'setOwnerACL', param)
        except CTERAException as error:
            logger.error('An error occurred attempting to update owner.')
            raise CTERAException('Owner modification failed.') from error


class Backups(BaseCommand):
    """ Backup Folder APIs """

    default = ['name', 'group', 'owner']

    def _get_entire_object(self, name):
        return self._core.api.get(f'/backups/{name}')

    def add(self, name, group, owner, xattr=True):
        """
        Create a new Backup Folder

        :param str name: Name of the backup folder
        :param str group: Folder Group to assign this folder to
        :param cterasdk.core.types.UserAccount owner: User account, the owner of the backup folder
        :param bool,optional xattr: Backup extended attributes, defaults to True
        """
        param = Object()
        param._classname = 'Backup'  # pylint: disable=protected-access
        param.name = name
        param.owner = self._core.users.get(owner, ['baseObjectRef']).baseObjectRef
        param.group = self._core.cloudfs.groups.get(group, ['baseObjectRef']).baseObjectRef
        param.enableBackupExtendedAttributes = xattr
        try:
            response = self._core.api.add('/backups', param)
            logger.info(
                'Backup folder created. %s',
                {'name': name, 'owner': param.owner, 'folder_group': group, 'xattr': xattr}
            )
            return response
        except CTERAException as error:
            logger.error(
                'Backup folder creation failed. %s',
                {'name': name, 'folder_group': group, 'owner': owner, 'xattr': xattr}
            )
            raise error

    def modify(self, current_name, new_name=None, new_owner=None, new_group=None, xattr=None):
        """
        Modify a Backup Folder

        :param str current_name: Current folder name
        :param cterasdk.core.types.UserAccount owner: User account, the owner of the folder
        :param str,optional new_name: New folder name
        :param cterasdk.core.types.UserAccount,optional new_owner: User account, the new owner of the folder
        :param str,optional new_group: Folder Group to assign this folder to
        :param bool,optional xattr: Backup extended attributes
        """
        param = self._get_entire_object(current_name)
        if new_name:
            param.name = new_name
        if new_owner:
            param.owner = self._core.users.get(new_owner, ['baseObjectRef']).baseObjectRef
        if new_group:
            param.group = self._core.cloudfs.groups.get(new_group, include=['baseObjectRef']).baseObjectRef
        if xattr:
            param.enableBackupExtendedAttributes = xattr
        try:
            response = self._core.api.put(f'/backups/{current_name}', param)
            logger.info('Backup folder updated. %s', {'name': current_name})
            return response
        except CTERAException as error:
            logger.error('Backup folder update failed. %s', {'name': current_name})
            raise error

    def all(self, include=None, list_filter=ListFilter.NonDeleted, user=None):
        """
        List Backup folders.

        :param str,optional include: List of fields to retrieve, defaults to ['name', 'group', 'owner']
        :param cterasdk.core.enum.ListFilter list_filter: Filter the list of Backup folders, defaults to non-deleted folders
        :param cterasdk.core.types.UserAccount user: Filter by backup folder owner
        :returns: Iterator for all Backup folders
        """
        include = union(include or [], Backups.default)
        builder = query.QueryParamBuilder().include(include)
        if list_filter != ListFilter.NonDeleted:
            builder.put('includeDeleted', True)
            if list_filter == ListFilter.Deleted:
                query_filter = query.FilterBuilder('isDeleted').eq(True)
                builder.addFilter(query_filter)
        if user:
            uid = self._core.users.get(user, ['uid']).uid
            builder.ownedBy(uid)
        param = builder.build()
        return query.iterator(self._core, '/backups', param)

    def delete(self, name):
        """
        Delete a Backup Folder

        :param str name: Name of the Backup Folder to delete
        """
        logger.info('Deleting Backup folder. %s', {'name': name})
        response = self._core.api.execute(f'/backups/{name}', 'delete')
        logger.info('Backup folder deleted. %s', {'name': name})
        return response


class Zones(BaseCommand):
    """
    Portal Zones APIs
    """

    name_attr = 'name'

    def get(self, name):
        """
        Get zone by name

        :param str name: The name of the zone to get
        :return: The requested zone
        """
        query_filter = query.FilterBuilder('name').eq(name)
        param = query.QueryParamBuilder().include_classname().startFrom(0).countLimit(1).addFilter(query_filter).orFilter(False).build()

        logger.info('Retrieving zone. %s', {'name': name})

        response = self._core.api.execute('', 'getZonesDisplayInfo', param)

        objects = response.objects
        if len(objects) < 1:
            logger.error('Zone not found. %s', {'name': name})
            raise CTERAException(f'Zone not found: {name}')

        zone = objects[0]
        logger.info('Zone found. %s', {'name': name, 'id': zone.zoneId})
        return zone

    def all(self, filters=None):
        """
        List Zones
        :param list[],optional filters: List of additional filters, defaults to None

        :return: Iterator for all Zones
        :rtype: cterasdk.lib.iterator.Iterator
        """
        builder = query.QueryParamBuilder().include_classname().startFrom(0).countLimit(25)
        filters = filters or []
        for query_filter in filters:
            builder.addFilter(query_filter)
        builder.orFilter((len(filters) > 1))
        param = builder.build()
        return query.iterator(self._core, '', param, 'getZonesDisplayInfo')

    def search(self, name):
        """
        Search for Zones by name
        :param str name: Search query

        :return: Iterator for all matching Zones
        :rtype: cterasdk.lib.iterator.Iterator
        """
        filters = [query.FilterBuilder(Zones.name_attr).like(name)]
        return self.all(filters)

    def add(self, name, policy_type=PolicyType.SELECT, description=None):
        """
        Add a new zone

        :param str name: The name of the new zone
        :param cterasdk.core.enum.PolicyType,optional policy_type:
         Policy type of the new zone, defaults to cterasdk.core.enum.PolicyType.SELECT
        :param str,optional description: The description of the new zone
        """
        param = self._zone_param(name, policy_type, description)

        logger.info('Adding zone. %s', {'name': name})

        response = self._core.api.execute('', 'addZone', param)
        try:
            self._process_response(response)
            logger.info('Zone added. %s', {'name': name})
        except CTERAException as error:
            logger.error('Zone creation failed. %s', {'rc': response.rc})
            raise error

    def delete(self, name):
        """
        Delete a zone

        :param str name: The name of the zone to delete
        """
        zone = self._core.cloudfs.zones.get(name)
        logger.info('Deleting zone. %s', {'zone': name})
        response = self._core.api.execute('', 'deleteZones', [zone.zoneId])
        if response == 'ok':
            logger.info('Zone deleted. %s', {'zone': name})

    def add_devices(self, name, device_names):
        """
        Add devices to a zone

        :param str name: The name of the zone to add devices to
        :param list[str] device_names: The names of the devices to add to the zone
        """
        zone = self._core.cloudfs.zones.get(name)
        portal_devices = devices.Devices(self._core).by_name(include=['uid'], names=device_names)
        info = self._zone_info(zone.zoneId)
        description = (info.description if hasattr(info, 'description') else None)

        param = self._zone_param(info.name, info.policyType, description, info.zoneId)
        for portal_device in portal_devices:
            param.delta.devicesDelta.added.append(portal_device.uid)

        logger.info('Adding devices to zone. %s', {'zone': info.name})

        try:
            self._save(param)
        except CTERAException as error:
            logger.error('Failed adding devices to zone.')
            raise CTERAException(f"Failed adding devices: [{', '.join(device_names)}] to zone: {name}") from error

    def add_folders(self, name, folder_finding_helpers):
        """
        Add the folders to the zone

        :param str name: The name of the zone
        :param list[cterasdk.core.types.CloudFSFolderFindingHelper] folder_finding_helpers: List of folder names and owners
        """
        zone = self._core.cloudfs.zones.get(name)
        folders = self._find_folders(folder_finding_helpers)
        info = self._zone_info(zone.zoneId)
        description = info.description if hasattr(info, 'description') else None
        param = self._zone_param(info.name, info.policyType, description, info.zoneId)
        param.delta.policyDelta = []

        for owner_id, folder_ids in folders.items():
            policyDelta = Object()
            policyDelta._classname = 'ZonePolicyDelta'  # pylint: disable=protected-access
            policyDelta.userUid = owner_id
            policyDelta.foldersDelta = Object()
            policyDelta.foldersDelta._classname = 'ZoneFolderDelta'  # pylint: disable=protected-access
            policyDelta.foldersDelta.added = copy.deepcopy(folder_ids)
            policyDelta.foldersDelta.removed = []

            param.delta.policyDelta.append(policyDelta)

        try:
            self._save(param)
        except CTERAException as error:
            logger.error('Failed adding folders to zone.')
            raise CTERAException(f'Failed adding folders to zone: {name}') from error

    def _zone_info(self, zid):
        logger.debug('Obtaining zone info. %s', {'id': zid})
        response = self._core.api.execute('', 'getZoneBasicInfo', zid)
        logger.debug('Obtained zone info. %s', {'id': zid})
        return response

    def _find_folders(self, folder_finding_helpers):
        folders = {}
        for folder_finding_helper in folder_finding_helpers:
            cloud_folder = CloudDrives(self._core).find(
                folder_finding_helper.name,
                folder_finding_helper.owner,
                include=['uid', 'owner']
            )
            folder_id = cloud_folder.uid
            owner_id = re.search("[1-9][0-9]*", cloud_folder.owner).group(0)

            if folders.get(owner_id) is None:
                folders[owner_id] = [folder_id]
            else:
                folders.get(owner_id).append(folder_id)

        return folders

    def _save(self, param):
        zone_name = param.basicInfo.name

        logger.debug('Applying changes to zone. %s', {'zone': param.basicInfo.name})

        response = self._core.api.execute('', 'saveZone', param)
        try:
            self._process_response(response)
        except CTERAException as error:
            logger.error('Failed applying changes to zone. %s', {'zone': zone_name, 'rc': response.rc})
            raise error

        logger.debug('Zone changes applied successfully. %s', {'zone': zone_name})

    @staticmethod
    def _process_response(response):
        if response.rc != 'OK':
            raise CTERAException(f'Zone creation failed: {response}')

    @staticmethod
    def _zone_param(name, policy_type, description=None, zid=None):
        param = Object()
        param._classname = "SaveZoneParam"  # pylint: disable=protected-access
        param.basicInfo = Object()
        param.basicInfo._classname = 'ZoneBasicInfo'  # pylint: disable=protected-access
        param.basicInfo.name = name
        param.basicInfo.policyType = policy_type
        param.basicInfo.description = description
        param.basicInfo.zoneId = zid
        param.delta = Object()
        param.delta._classname = 'ZoneDelta'  # pylint: disable=protected-access
        param.delta.devicesDelta = Object()
        param.delta.devicesDelta._classname = 'ZoneDeviceDelta'  # pylint: disable=protected-access
        param.delta.devicesDelta.added = []
        param.delta.devicesDelta.removed = []
        return param


class Exports(BaseCommand):
    """ S3 Exports APIs """

    def get(self, name):
        """
        Get Buckets

        :param str name: Bucket name
        """
        return self._core.api.get(f'/buckets/{name}')

    def get_endpoint(self, name):
        """
        Get Endpoint

        :param str name: Bucket name
        :returns: Bucket endpoint
        :rtype: str
        """
        return self.get(name).url

    def all(self):
        """
        List Buckets
        """
        param = query.QueryParamBuilder().startFrom(0).countLimit(25).orFilter(True).build()
        return query.iterator(self._core, '/buckets', param)

    def add(self, name, drive_name, drive_owner, description=None):
        """
        Add Bucket

        :param str name: Bucket name
        :param str drive_name: Cloud Drive Folder name
        :param str drive_owner: Cloud Drive Folder owner
        :param str,optional description: Bucket description
        """
        param = Object()
        param._classname = 'Bucket'  # pylint: disable=protected-access
        param.description = description
        param.name = name
        param.cloudDrive = self._core.cloudfs.drives.find(drive_name, drive_owner, include=['baseObjectRef']).baseObjectRef
        logger.info('Adding Bucket. %s', {'name': name})
        response = self._core.api.add('/buckets', param)
        logger.info('Bucket Added. %s', {'name': name})
        return response

    def modify(self, name, description):
        """
        Modify Bucket

        :param str name: Bucket name
        :param str description: Bucket description
        """

        bucket = self.get(name)
        bucket.description = description
        logger.info("Modifying Bucket. %s", {'name': name})
        response = self._core.api.put(f'/buckets/{name}', bucket)
        logger.info("Bucket modified. %s", {'name': name})
        return response

    def delete(self, name):
        """
        Remove Bucket

        :param str name: Bucket name
        """
        logger.info('Deleting Bucket. %s', {'name': name})
        response = self._core.api.delete(f'/buckets/{name}')
        logger.info('Bucket deleted. %s', {'name': name})
        return response


class Locks(BaseCommand):

    def all(self):
        """
        List Locked Files

        :returns: Iterator for all locked files
        """
        builder = query.QueryParamBuilder().include_classname()
        param = builder.build()
        return query.iterator(self._core, '', param, 'getLockedFilesInfo')



import logging
import time

from ..exceptions.transport import HTTPError
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Power(BaseCommand):
    """ Edge Filer Power APIs """

    def reboot(self, wait=False):
        """
        Reboot the Gateway

        :param bool,optional wait: Wait for reboot to complete, defaults to False
        """
        logger.info("Rebooting Edge Filer. %s", {'host': self._edge.host()})
        self._edge.api.execute("/status/device", "reboot", None)
        if wait:
            Boot(self._edge).wait()

    def shutdown(self):
        """ Shutdown the Edge Filer"""
        self._edge.api.execute("/status/device", "poweroff", None)

    def reset(self, wait=False):
        """
        Reset the Edge Filer setting

        :param bool,optional wait: Wait for reset to complete, defaults to False
        """
        self._edge.api.execute("/status/device", "reset2default", None)
        logger.info("Resetting Edge Filer to default settings. %s", {'host': self._edge.host()})
        if wait:
            Boot(self._edge).wait()


class Boot:

    def __init__(self, edge, retries=60, seconds=5):
        self._edge = edge
        self._retries = retries
        self._seconds = seconds
        self._attempt = 0

    def wait(self):
        while True:
            try:
                self._increment()
                logger.debug("Status check, (try %s)", self._attempt + 1)
                self._edge.test()
                logger.info("Edge Filer is up and running.")
                break
            except ConnectionError:
                logger.debug('Connection error while checking status.')
            except TimeoutError:
                logger.debug('Status check timed out.')
            except HTTPError as e:
                logger.debug("Status check failed with HTTP %s: %s", e.code, e.name)

    def _increment(self):
        self._attempt = self._attempt + 1
        if self._attempt >= self._retries:
            self._unreachable()
        logger.debug("Try %s failed; Sleeping for %s second(s).", self._attempt, self._seconds)
        time.sleep(self._seconds)

    def _unreachable(self):
        host = self._edge.host()
        port = self._edge.port()
        logger.error("Connection timed out after retries. %s", {'host': host, 'port': port})
        raise ConnectionError(f'Connection to {host}:{port} timed out after retries.')



import aiohttp
from ...audit import postman


def tracer():

    async def on_request_start(session, ctx, params):
        # pylint: disable=unused-argument
        """
        :param aiohttp.ClientSession session: Session.
        :param cterasdk.transcriber.postman.Command ctx: Command.
        """
        ctx.stream = postman.BodyStream()
        ctx.request = postman.Request(
            params.method,
            postman.URL(
                str(params.url),
                params.url.scheme,
                params.url.host.split('.'),
                params.url.port,
                params.url.path.split('/')[1:],
                dict(params.url.query)
            )
        )

    async def on_request_headers_sent(session, ctx, params):
        # pylint: disable=unused-argument
        """
        :param aiohttp.ClientSession session: Session.
        :param cterasdk.transcriber.postman.Command ctx: Command.
        """
        ctx.request.request_headers({k: v for k, v in params.headers.items() if k in [
            'Cookie',
            'Authorization',
            'Content-Type',
            'Depth',
            'Overwrite',
            'Destination'
        ]})

    async def on_request_chunk_sent(session, ctx, params):
        # pylint: disable=unused-argument
        """
        :param aiohttp.ClientSession session: Session.
        :param cterasdk.transcriber.postman.Command ctx: Command.
        """
        ctx.stream.append(params.chunk)

    async def on_request_end(session, ctx, params):
        # pylint: disable=unused-argument
        """
        :param aiohttp.ClientSession session: Session.
        :param cterasdk.transcriber.postman.Command ctx: Command.
        """
        body = ctx.stream.deserialize(params.response.request_info.headers.get('Content-Type', None))
        if body:
            ctx.request.request_body(body)
        postman.Collection.instance().add(ctx.request)

    conf = aiohttp.TraceConfig()
    conf.on_request_start.append(on_request_start)
    conf.on_request_headers_sent.append(on_request_headers_sent)
    conf.on_request_chunk_sent.append(on_request_chunk_sent)
    conf.on_request_end.append(on_request_end)

    return conf



import logging
from ....cio.common import encode_request_parameter
from ....cio import core as fs
from ....exceptions.io import ResourceNotFoundError, ResourceExistsError
from .. import query


logger = logging.getLogger('cterasdk.core')


async def listdir(core, path, depth=None, include_deleted=False, search_criteria=None, limit=None):
    with fs.fetch_resources(path, depth, include_deleted, search_criteria, limit) as param:
        if param.depth > 0:
            return query.iterator(core, '', param, 'fetchResources', callback_response=fs.FetchResourcesResponse)
        return await core.v1.api.execute('', 'fetchResources', param)


async def exists(core, path):
    present, *_ = await metadata(core, path, suppress_error=True)
    return present


async def metadata(core, path, suppress_error=False):
    """
    Get item metadata.

    :returns: A tuple indicating if a file exists, and its metadata
    """
    response = await listdir(core, path, 0)
    if response.root is None:
        if not suppress_error:
            raise ResourceNotFoundError(path.absolute)
        return False, None
    return True, response.root


async def versions(core, path):
    with fs.versions(path):
        return await core.v1.api.execute('', 'listSnapshots', path.absolute)


async def walk(core, scope, path, include_deleted=False):
    target = fs.CorePath.instance(scope, path)
    await ensure_directory(core, target)
    paths = [fs.CorePath.instance(scope, path)]
    while len(paths) > 0:
        path = paths.pop(0)
        entries = await listdir(core, path, include_deleted=include_deleted)
        async for e in entries:
            if e.isFolder:
                paths.append(fs.CorePath.instance(scope, e))
            yield e


async def mkdir(core, path):
    with fs.makedir(path) as param:
        response = await core.v1.api.execute('', 'makeCollection', param)
    fs.accept_error(response)


async def makedirs(core, path):
    directories = path.parts
    for i in range(1, len(directories) + 1):
        path = fs.CorePath.instance(path.scope, '/'.join(directories[:i]))
        try:
            await mkdir(core, path)
        except ResourceExistsError:
            logger.debug('Resource already exists: %s', path.reference.as_posix())


async def rename(core, path, name):
    with fs.rename(path, name) as param:
        return await core.v1.api.execute('', 'moveResources', param)


async def remove(core, *paths):
    with fs.delete(*paths) as param:
        return await core.v1.api.execute('', 'deleteResources', param)


async def recover(core, *paths):
    with fs.recover(*paths) as param:
        return await core.v1.api.execute('', 'restoreResources', param)


async def copy(core, *paths, destination=None, resolver=None, cursor=None):
    with fs.copy(*paths, destination=destination, resolver=resolver, cursor=cursor) as param:
        return await core.v1.api.execute('', 'copyResources', param)


async def move(core, *paths, destination=None, resolver=None, cursor=None):
    with fs.move(*paths, destination=destination, resolver=resolver, cursor=cursor) as param:
        return await core.v1.api.execute('', 'moveResources', param)


async def ensure_directory(core, directory, suppress_error=False):
    present, resource = await metadata(core, directory, suppress_error=True)
    fs.ensure_directory(present, resource, directory, suppress_error)
    return resource.isFolder if present else False, resource


def handle(path):
    """
    Create function to retrieve file handle.

    :param cterasdk.cio.edge.CorePath path: Path to file.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    async def wrapper(core):
        """
        Get file handle.

        :param cterasdk.objects.synchronous.core.Portal core: Portal object.
        """
        with fs.handle(path) as param:
            return await core.io.download(param)
    return wrapper


def handle_many(directory, *objects):
    """
    Create function to retrieve zip archive

    :param cterasdk.cio.edge.CorePath directory: Path to directory.
    :param args objects: List of files and folders.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    async def wrapper(core):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.core.Portal core: Portal object.
        :param str name: File name.
        :param object handle: File handle.
        """
        with fs.handle_many(directory, objects) as param:
            _, resource = await ensure_directory(core, directory)
            return await core.io.download_zip(str(resource.cloudFolderInfo.uid), encode_request_parameter(param))
    return wrapper


async def _validate_destination(core, name, destination):
    is_dir, resource = await ensure_directory(core, destination, suppress_error=True)
    if not is_dir:
        is_dir, resource = await ensure_directory(core, destination.parent)
        fs.ensure_writeable(resource, destination.parent)
        return resource.cloudFolderInfo.uid, destination.name, destination.parent
    fs.ensure_writeable(resource, destination)
    return resource.cloudFolderInfo.uid, name, destination


def upload(name, size, destination, fd):
    """
    Create upload function

    :param str name: File name.
    :param cterasdk.cio.core.CorePath destination: Path to directory.
    :param object fd: File handle.
    :returns: Callable function to start the upload.
    :rtype: callable
    """
    fs.destination_prerequisite_conditions(destination, name)

    async def wrapper(core):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.core.Portal core: Portal object.
        """
        uid, filename, directory = await _validate_destination(core, name, destination)
        with fs.upload(core, filename, directory, size, fd) as param:
            return fs.validate_transfer_success(await core.io.upload(str(uid), param), destination.join(name).reference.as_posix())
    return wrapper


async def public_link(core, path, access, expire_in):
    with fs.public_link(path, access, expire_in) as param:
        response = await core.v1.api.execute('', 'createShare', param)
    return response.publicLink



import logging

from . import query
from ..lib.tasks import AwaitableEdgeTask
from ..common import Object
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Tasks(BaseCommand):
    """ Edge Filer Background Task APIs """

    def awaitable_task(self, ref):
        return AwaitableEdgeTask(self._edge, ref)

    def wait(self, ref, timeout=None, poll_interval=None):
        awaitable_task = AwaitableEdgeTask(self._edge, ref)
        return awaitable_task.wait(timeout, poll_interval)

    def running(self):
        """
        Get all running background tasks
        """
        return self._query('status', 'running')

    def by_name(self, name):
        """
        Get background tasks by name

        :param str name: Task name
        """
        return self._query('name', name)

    def _query(self, key, value):
        param = Object()
        param.key = key
        param.value = value
        return query.iterator(self._edge, '/proc/bgtasks', param)



import logging

from .types import AlertSettings
from ..common import Object
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Mail(BaseCommand):
    """
    Edge Filer Mail Server configuration APIs

    :ivar cterasdk.edge.mail.alerts alerts: Object holding the Edge Filer e-mail alert APIs
    """

    def __init__(self, edge):
        super().__init__(edge)
        self.alerts = Alerts(self._edge)

    def enable(self, smtp_server, port=25, username=None, password=None, use_tls=True,
               sender=None, recipients=None, min_severity=None):
        """
        Enable e-mail delivery using a custom SMTP server

        :param str smtp_server: Address of the SMTP Server
        :param int,optional port: The listening port of the SMTP Server, defaults to 25
        :param str,optional username: The user name of the SMTP Server, defaults to None
        :param str,optional password: The password of the SMTP Server, defaults to None
        :param bool,optional use_tls: Use TLS when connecting to the SMTP Server, defaults to True
        :param str,optional sender: Sender e-mail address
        :param list[str],optional recipients: List of e-mail recipients
        :param cterasdk.edge.enum.Severity severity: Minimum severity level to trigger email alerts
        """
        settings = self._edge.api.get('/config/logging/alert')
        settings.useCustomServer = True
        settings.SMTPServer = smtp_server

        if settings.port != port:
            settings.port = port

        if username is not None and password is not None:
            settings.useAuth = True
            settings.auth = Object()
            settings.auth._classname = 'AuthSettings'  # pylint: disable=protected-access
            settings.auth.username = username
            settings.auth.password = password

        if settings.useTLS != use_tls:
            settings.useTLS = use_tls

        if sender is not None:
            setattr(settings, 'from', sender)

        if recipients is not None:
            settings.emails = recipients

        if min_severity is not None:
            settings.minSeverity = min_severity

        logger.info('Enabling mail server.')

        self._edge.api.put('/config/logging/alert', settings)

        logger.info(
            'Updated mail server settings. %s',
            {'SMTPServer': smtp_server, 'port': port, 'username': username, 'useTLS': use_tls}
        )

    def disable(self):
        """ Disable e-mail delivery using a custom SMTP server """
        logger.info('Disabling mail server.')
        self._edge.api.put('/config/logging/alert/useCustomServer', False)
        logger.info('Mail server disabled.')


class Alerts(BaseCommand):

    def get(self):
        """
        Get Alert Settings

        :returns: Alert settings
        :rtype: cterasdk.edge.types.AlertSettings
        """
        return AlertSettings.from_server_object(self._edge.api.get('/config/logging/alert/specificAlerts'))

    def modify(self, alerts):
        """
        Modify Alert Settings

        :param cterasdk.edge.types.AlertSettings alerts: Alert Settings
        """
        return AlertSettings.from_server_object(self._edge.api.put('/config/logging/alert/specificAlerts', alerts.to_server_object()))



import logging


logger = logging.getLogger('cterasdk.direct')


class BaseCredentials:
    """Base Credentials for CTERA Direct IO"""


class KeyPair(BaseCredentials):
    """Access and Secret Key Pair"""

    def __init__(self, access_key_id, secret_access_key):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key


class Bearer(BaseCredentials):
    """Bearer Token"""

    def __init__(self, bearer):
        self.bearer = bearer


def create_bearer_token(credentials):
    """
    Create Authorization Header.

    :param cterasdk.direct.credentials.BaseCredentials credentials: Credentials
    :returns: Authorization header as a dictionary.
    :rtype: dict
    """
    token = None

    if isinstance(credentials, Bearer):
        logger.debug('Initializing client using Bearer token')
        token = f'Bearer {credentials.bearer}'

    elif isinstance(credentials, KeyPair):
        logger.debug('Initializing client using Key Pair.')
        token = f'Bearer {credentials.access_key_id}'

    return token



import json
import logging
from pathlib import Path
import yaml
from .convert import fromjsonstr


logger = logging.getLogger('cterasdk')


settings = Path(__file__).parent.absolute().joinpath('settings.yml')
try:
    with open(settings, 'r', encoding='utf-8') as f:
        settings = fromjsonstr(json.dumps(yaml.safe_load(f)))
    core, edge, drive, io, audit = settings.core, settings.edge, settings.drive, settings.io, settings.audit
except FileNotFoundError:
    logger.fatal("Configuration file 'settings.yml' not found. Please check your installation and try again.")
    raise



from .base import CTERAException


class NotificationsError(CTERAException):
    """
    Notifications error

    :ivar list[int] cloudfolders: List of cloudfolders
    :ivar str cursor: Cursor
    """

    def __init__(self, cloudfolders, cursor):
        super().__init__('An error occurred while trying to retrieve notifications.')
        self.cloudfolders = cloudfolders
        self.cursor = cursor


class AncestorsError(CTERAException):
    """
    Ancestors Error

    :ivar int folder_id: Cloud Drive folder unique identifer
    :ivar str guid: File GUID
    """
    def __init__(self, folder_id, guid):
        super().__init__(f'Could not retrieve ancestors for: {folder_id}:{guid}')
        self.folder_id = folder_id
        self.guid = guid
